/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/fb.h>

#include "viafbdev.h"
#include "tv.h"
#include "share.h"
#include "chip.h"
#include "debug.h"
#include "tblIntegratedTV.h"
#include "IntegratedTV.h"
#include "ioctl.h"

extern struct tv_setting_information    tv_setting_info;
extern struct chip_information          chip_info;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
extern struct viafb_info    viafbinfo;
#else
extern struct viafb_par     parinfo;
#endif

extern inline void write_reg(u8 index, u16 io_port, u8 data);
extern inline u8 read_reg(int io_port, u8 index);
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern void enable_second_display_channel(void);
extern void plus_vck_to_iga1_timing(unsigned long h_total);
extern void delays(int count);
extern void load_crtc_timing(struct display_timing device_timing, int set_iga);

bool g_first_time_to_sense = TRUE;

void enable_integrated_tv_shadow_reg(void)
{
    int i;

    /* Enable integrated TV shadow register. */
    /* Only for A0000 now, we may modify this if need other TV shadow memory. */
    write_reg(SR52, VIASR, 0x01);

    /* Wait data ready. */
    for (i=0; i<10; i++)
    {
        if (read_reg(VIASR, SR52) & BIT0)
        {
            break;
        }
    }
}


void disable_integrated_tv_shadow_reg(void)
{
    write_reg(SR52, VIASR, 0x00);
}  


void write_integrated_tv_reg8(unsigned char* mmio_base, u8 offset, u8 data)
{
    *(volatile u8 *)(mmio_base + offset) = data;
    DEBUG_MSG(KERN_INFO "Write Integrated TV Reg (byte): Offset 0x%2x, Value 0x%2x!\n", offset, data);
}


void write_integrated_tv_reg32(unsigned char* mmio_base, u8 offset, u32 data)
{
    *(volatile u32 *)(mmio_base + offset) = data;
    DEBUG_MSG(KERN_INFO "Write Integrated TV Reg: Offset 0x%2x, Value 0x%8x!\n", offset, data);
}


void write_integrated_tv_reg_mask32(unsigned char* mmio_base, u8 offset, u32 mask, u32 data)
{
    u32 old_data, new_data;

    old_data = *(volatile u32 *)(mmio_base + offset);
    new_data = (data & mask) | (old_data & (~mask));
    *(volatile u32 *)(mmio_base + offset) = new_data;

    DEBUG_MSG(KERN_INFO "Write Integrated TV Reg Mask: Offset 0x%2x, Value 0x%8x!\n", offset, new_data);
}


u32 read_integrated_tv_reg32(unsigned char* mmio_base, u8 offset)
{
    u32 data;

    data = *(volatile u32 *)(mmio_base + offset);

    return data;
}


void write_integrated_tv_general_table(unsigned short *tv_reg_offset_table, unsigned long *tv_reg_table)
{
    int i = 0;
    unsigned char* mmio_base;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    mmio_base = (unsigned char*)viafbinfo.integrated_tv_map_base;
#else
    mmio_base = (unsigned char*)parinfo.integrated_tv_map_base;
#endif

    while(tv_reg_offset_table[i] != 0xFFFF)
    {
        write_integrated_tv_reg32(mmio_base, GET_LOW_BYTE(tv_reg_offset_table[i]), tv_reg_table[i]);
        i++;
    }
}


void write_integrated_tv_patch_table(u16 *tv_reg_table)
{
    int i = 0;
    unsigned char* mmio_base;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    mmio_base = (unsigned char*)viafbinfo.integrated_tv_map_base;
#else
    mmio_base = (unsigned char*)parinfo.integrated_tv_map_base;
#endif
    
    while(tv_reg_table[i] != 0xFFFF)
    {
        write_integrated_tv_reg8(mmio_base, GET_LOW_BYTE(tv_reg_table[i]), GET_HIGH_BYTE(tv_reg_table[i]));
        i++;
    }
}


u32 access_integrated_tv_regs(int reg_id, u32 value, int cmd)
{
    u8 offset;
    u32 mask, set_data, get_data, shift_bits, tmp1, tmp2;
    unsigned char* mmio_base;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    mmio_base = (unsigned char*)viafbinfo.integrated_tv_map_base;
#else
    mmio_base = (unsigned char*)parinfo.integrated_tv_map_base;
#endif

    /* Note: It's better to call enable_integrated_tv_shadow_reg() to enable integrated TV shadow register */
    /* before access the following registers, or it may fail to access them. */
    
    switch (reg_id)
    {
        case ITV_REG_ID_PK_ENABLE:
        {
            /* 00[0] */
            offset = 0x00;
            mask = 0x00000001;
            shift_bits = 0;

            break;
        }

        case ITV_REG_ID_SRC_SEL:
        {
            /* 00[18] */
            offset = 0x00;
            mask = 0x00040000;
            shift_bits = 18;
            
            break;
        }

        case ITV_REG_ID_SW_RESET:
        {
            /* 00[20] */
            offset = 0x00;
            mask = 0x00100000;
            shift_bits = 20;
            
            break;
        }

        case ITV_REG_ID_SYNC_BUF_MODE:
        {
            /* 00[30] */
            offset = 0x00;
            mask = 0x40000000;
            shift_bits = 30;
            
            break;
        }

        case ITV_REG_ID_SD_ENABLE:
        {
            /* 00[31] */
            offset = 0x00;
            mask = 0x80000000;
            shift_bits = 31;
            
            break;
        }

        case ITV_REG_ID_SRC_VSIZE:
        {
            /* 18[10:0] */
            offset = 0x18;
            mask = 0x000007FF;
            shift_bits = 0;
            
            break;
        }

        case ITV_REG_ID_SRC_HSIZE:
        {
            /* 18[21:11] */
            offset = 0x18;
            mask = 0x003FF800;
            shift_bits = 11;
            
            break;
        }

        case ITV_REG_ID_DE_FLK_THRESHOLD:
        {
            /* 20[31:27] + 20[15:11] */
            offset = 0x20;
                
            if (cmd == ITV_CMD_WRITE)
            {
                tmp1 = value & 0x1F;        /* bit4-bit0 */
                tmp2 = (value >> 5) & 0x1F; /* bit9-bit5 */
                set_data = (tmp2 << 27) | (tmp1 << 11);
                write_integrated_tv_reg_mask32(mmio_base, offset, 0xF800F800, set_data);
                
                return 0;  /* 0 means nothing, just for return. */
            }
            else
            {
                get_data = read_integrated_tv_reg32(mmio_base, offset);
                tmp1 = (get_data & 0x0000F800) >> 11;   /* bit4-bit0 */
                tmp2 = (get_data & 0xF8000000) >> 27;   /* bit9-bit5 */
                get_data = (tmp2 << 5) | tmp1;

                return get_data;
            }
        }
        
        case ITV_REG_ID_SCL_VFAC:
        {
            /* 24[12:0] */
            offset = 0x24;
            mask = 0x00001FFF;
            shift_bits = 0;
            
            break;
        }

        case ITV_REG_ID_DE_FLK_NUM:
        {
            /* 24[13] */
            offset = 0x24;
            mask = 0x00002000;
            shift_bits = 13;
            
            break;
        }
        
        case ITV_REG_ID_DF_WT_SEL:
        {
            /* 24[14] */
            offset = 0x24;
            mask = 0x00004000;
            shift_bits = 14;
            
            break;
        }

        case ITV_REG_ID_DF_TYPE:
        {
            /* 24[15] */
            offset = 0x24;
            mask = 0x00008000;
            shift_bits = 15;
            
            break;
        }
        
        case ITV_REG_ID_DE_FLICKER_ENABLE:
        {
            /* 24[29] */
            offset = 0x24;
            mask = 0x20000000;
            shift_bits = 29;
            
            break;
        }
        
        case ITV_REG_ID_SCL_HFAC:
        {
            /* 24[28:16] */
            offset = 0x24;
            mask = 0x1FFF0000;
            shift_bits = 16;
            
            break;
        }

        case ITV_REG_ID_DST_VSIZE:
        {
            /* 28[10:0] */
            offset = 0x28;
            mask = 0x000007FF;
            shift_bits = 0;
            
            break;
        }

        case ITV_REG_ID_DST_HSIZE:
        {
            /* 28[21:11] */
            offset = 0x28;
            mask = 0x003FF800;
            shift_bits = 11;
            
            break;
        }

        case ITV_REG_ID_TV_ENABLE:
        {
            /* 38[0] */
            offset = 0x38;
            mask = 0x00000001;
            shift_bits = 0;

            break;
        }

        case ITV_REG_ID_OUTPUT_SEL:
        {
            /* 38[29:28] */
            offset = 0x38;
            mask = 0x30000000;
            shift_bits = 28;

            break;
        }

        case ITV_REG_ID_TV_SENSE:
        {
            /* 3C[15:14] */
            offset = 0x3C;
            mask = 0x0000C000;
            shift_bits = 14;
            
            break;
        }
        
        case ITV_REG_ID_TV_SENSE_DATA:
        {
            /* 3C[31:24] */
            offset = 0x3C;
            mask = 0xFF000000;
            shift_bits = 24;
            
            break;
        }
        
        case ITV_REG_ID_DST_HSD:
        {
            /* 48[30:21] */
            offset = 0x48;
            mask = 0x7FE00000;
            shift_bits = 21;

            break;
        }

        case ITV_REG_ID_DST_VSD:
        {
            /* 4C[31:23] */
            offset = 0x4C;
            mask = 0xFF800000;
            shift_bits = 23;

            break;
        }

        case ITV_REG_ID_SATURATION:
        {
            /* 50[8:0] + 50[24:16] */
            offset = 0x50;
                
            if (cmd == ITV_CMD_WRITE)
            {
                tmp1 = value & 0x1FF;        /* bit8-bit0 */
                tmp2 = (value >> 9) & 0x1FF; /* bit24-bit16 */
                set_data = ((tmp2 << 16) | (tmp1));
                write_integrated_tv_reg_mask32(mmio_base, offset, 0x01FF01FF, set_data);
                
                return 0;  /* 0 means nothing, just for return. */
            }
            else
            {
                get_data = read_integrated_tv_reg32(mmio_base, offset);
                tmp1 = (get_data & 0x000001FF);   /* bit8-bit0 */
                tmp2 = (get_data & 0x01FF0000) >> 16;   /* bit24-bit16 */
                get_data = (tmp2 << 9) | tmp1;

                return get_data;
            }
        }

        case ITV_REG_ID_CONTRAST:
        {
            /* 54[8:0] */
            offset = 0x54;
            mask = 0x000001FF;
            shift_bits = 0;
            break;
        }
        
        case ITV_REG_ID_BLACK_LEVEL:
        {
            /* 54[30:23] */
            offset = 0x54;
            mask = 0x7F800000;
            shift_bits = 23;
            break;
        }

        case ITV_REG_ID_HUE:
        {
            /* 60[10:0] */
            offset = 0x60;
            mask = 0x000007FF;
            shift_bits = 0;
            break;
        }

        case ITV_REG_ID_PRI_DAC_SEL:
        {
            /* 6C[12] */
            offset = 0x6C;
            mask = 0x00001000;
            shift_bits = 12;
            
            break;
        }

        case ITV_REG_ID_TV_SENSE_STATUS:
        {
            /* E4[2] */
            offset = 0xE4;
            mask = 0x00000004;
            shift_bits = 2;
            
            break;
        }

        case ITV_REG_ID_PBPR_BLANK_LEVEL:
        {
            /* E8[8:0] */
            offset = 0xE8;
            mask = 0x000000FF;
            shift_bits = 0; 
            break;
        }

        case ITV_REG_ID_TV_SENSE_RGB_STATUS:
        {
            /* E8[30:28] */
            offset = 0xE8;
            mask = 0x70000000;
            shift_bits = 28;
            
            break;
        }
        
        default:
        {
            return 0;  /* 0 means nothing, just for return. */
        }
    }

    if (cmd == ITV_CMD_WRITE)
    {
        set_data = value << shift_bits;
        write_integrated_tv_reg_mask32(mmio_base, offset, mask, set_data);
        
        return 0;  /* 0 means nothing, just for return. */
    }
    else
    {
        get_data = read_integrated_tv_reg32(mmio_base, offset);
        get_data = (get_data & mask) >> shift_bits;

        return get_data;
    }
}


void set_integrated_tv_output_signal(void)
{
    u8  sr5e;
    
    /* HDTV only can use RGB or YPbPr output. */
    if ((tv_setting_info.system == TVTYPE_480P) || (tv_setting_info.system == TVTYPE_576P) ||
        (tv_setting_info.system == TVTYPE_720P) || (tv_setting_info.system == TVTYPE_1080I))
        
    {
        if ((tv_setting_info.out_signal != TV_OUTPUT_RGB) &&
            (tv_setting_info.out_signal != TV_OUTPUT_YPBPR))
        {
            tv_setting_info.out_signal = TV_OUTPUT_RGB;
        }
    }
    
    DEBUG_MSG(KERN_INFO "set_integrated_tv_output_signal: TVOutput = %d\n", tv_setting_info.out_signal);

    /* We could separate Composite and S-Video by controlling DAC R/G/B. */
    sr5e = read_reg(VIASR, SR5E) & 0xF1;    /* Turn on DAC R/G/B. */

    switch (tv_setting_info.out_signal)
    {
        case TV_OUTPUT_COMPOSITE:        
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_SC, ITV_CMD_WRITE);

            /* Turn off DAC R and G, turn on DAC B. */
            write_reg(SR5E, VIASR, sr5e | (BIT1+BIT2));
            break;
        }

        case TV_OUTPUT_SVIDEO:
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_SC, ITV_CMD_WRITE);

            /* Turn on DAC R and G, turn off DAC B. */
            write_reg(SR5E, VIASR, sr5e | BIT3);
            break;
        }

        case TV_OUTPUT_COMPOSITE_SVIDEO:
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_SC, ITV_CMD_WRITE);

            /* Turn on DAC R/G/B. */
            write_reg(SR5E, VIASR, sr5e);
            break;
        }
        
        case TV_OUTPUT_RGB:
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_RGB, ITV_CMD_WRITE);

            /* Turn on DAC R/G/B. */
            write_reg(SR5E, VIASR, sr5e);
            break;
        }
            
        case TV_OUTPUT_YPBPR:
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_YPbPr, ITV_CMD_WRITE);

            /* Turn on DAC R/G/B. */
            write_reg(SR5E, VIASR, sr5e);
            break;
        }
            
        default:
        {
            access_integrated_tv_regs(ITV_REG_ID_OUTPUT_SEL, ITV_VALUE_OUTPUT_SC, ITV_CMD_WRITE);

            /* Turn off DAC R and G, turn on DAC B. */
            write_reg(SR5E, VIASR, sr5e | (BIT1+BIT2));
            break;
        }
    }
}


void write_gfx_crtc_timing_for_tv(unsigned long *tv_crtc_table)
{
    int i;
    struct display_timing  tv_crtc_reg;
    
    for (i=0; i< NUM_CRTC_TIMING; i++)
    {
        switch(i)
        {
            case H_TOTAL_INDEX:
                tv_crtc_reg.hor_total = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Total: %d!\n", tv_crtc_reg.hor_total);
                break;
                
            case H_ADDR_INDEX:
                tv_crtc_reg.hor_addr = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Addr: %d!\n", tv_crtc_reg.hor_addr);
                break;
                
            case H_BLANK_START_INDEX:
                tv_crtc_reg.hor_blank_start = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Blank_Start: %d!\n", tv_crtc_reg.hor_blank_start);
                break;
                
            case H_BLANK_END_INDEX:
                tv_crtc_reg.hor_blank_end = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Blank_End: %d!\n", tv_crtc_reg.hor_blank_end);
                break;
                
            case H_SYNC_START_INDEX:
                tv_crtc_reg.hor_sync_start = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Sync_Start: %d!\n", tv_crtc_reg.hor_sync_start);
                break;
                
            case H_SYNC_END_INDEX:
                tv_crtc_reg.hor_sync_end = tv_crtc_table[i];
                DEBUG_MSG(KERN_INFO "H_Sync_End: %d!\n", tv_crtc_reg.hor_sync_end);
                break;
                
            case V_TOTAL_INDEX:
                tv_crtc_reg.ver_total = tv_crtc_table[i];
                break;
                
            case V_ADDR_INDEX:
                tv_crtc_reg.ver_addr = tv_crtc_table[i];
                break;
                
            case V_BLANK_START_INDEX:
                tv_crtc_reg.ver_blank_start = tv_crtc_table[i];
                break;
                
            case V_BLANK_END_INDEX:
                tv_crtc_reg.ver_blank_end = tv_crtc_table[i];
                break;
                
            case V_SYNC_START_INDEX:
                tv_crtc_reg.ver_sync_start = tv_crtc_table[i];
                break;
                
            case V_SYNC_END_INDEX:
                tv_crtc_reg.ver_sync_end = tv_crtc_table[i];
                break;
                
        }
    }

    load_crtc_timing(tv_crtc_reg, tv_setting_info.iga_path);
}


void update_scaling_factor(void)
{
    u32 dst_hsize, dst_vsize, src_hsize, src_vsize, scal_factor = 0;

    dst_hsize = access_integrated_tv_regs(ITV_REG_ID_DST_HSIZE, 0, ITV_CMD_READ) + 1;
    dst_vsize = access_integrated_tv_regs(ITV_REG_ID_DST_VSIZE, 0, ITV_CMD_READ) + 1;
    src_hsize = access_integrated_tv_regs(ITV_REG_ID_SRC_HSIZE, 0, ITV_CMD_READ) + 1;
    src_vsize = access_integrated_tv_regs(ITV_REG_ID_SRC_VSIZE, 0, ITV_CMD_READ) + 1;

    if ((tv_setting_info.system == TVTYPE_480P) ||
        (tv_setting_info.system == TVTYPE_576P) ||
        (tv_setting_info.system == TVTYPE_1080I))
    {
        scal_factor = (src_hsize * 2048) / dst_hsize;
    }
    else if (tv_setting_info.system == TVTYPE_720P)
    {
        /* The two modes 720x480 and 720x576 have alignment issue (alignment to 4), 
           so the value of SRC_LINE_PITCH, DST_DQW_NUM and DST_LINE_PITCH
           we set are greater than the desired ones, and this may cause the problem about 
           having garbage line on the right side, so we reduce the scaling factor to avoid
           this problem. */
        if (((src_hsize == 720) && (src_vsize == 480)) ||
            ((src_hsize == 720) && (src_vsize == 576)))
        {
            scal_factor = ((src_hsize * 2048) / dst_hsize) - 3;
        }
        else
        {
            scal_factor = (src_hsize * 2048) / dst_hsize;
        }
    }
    else
    {
        scal_factor = (src_hsize * 2048) / (dst_hsize * 2);
    }
    
    access_integrated_tv_regs(ITV_REG_ID_SCL_HFAC, scal_factor, ITV_CMD_WRITE);

    scal_factor = (src_vsize * 2048) / dst_vsize;
    access_integrated_tv_regs(ITV_REG_ID_SCL_VFAC, scal_factor, ITV_CMD_WRITE);
}


void set_hdtv_patch_table(int mode_index)
{
    switch(tv_setting_info.system)
    {
        case TVTYPE_480P:
            write_integrated_tv_patch_table(ITV_HDTV_PATCH_TABLE[mode_index].std_480p);
            break;
            
        case TVTYPE_576P:
            write_integrated_tv_patch_table(ITV_HDTV_PATCH_TABLE[mode_index].std_576p);
            break;
            
        case TVTYPE_720P:
            write_integrated_tv_patch_table(ITV_HDTV_PATCH_TABLE[mode_index].std_720p);
            break;
            
        case TVTYPE_1080I:
            write_integrated_tv_patch_table(ITV_HDTV_PATCH_TABLE[mode_index].std_1080i);
            break;

        default:
            break;
    }
}


void set_tv_output_patch_table(void)
{
    if (tv_setting_info.out_signal == TV_OUTPUT_RGB)
    {
        if (tv_setting_info.system == TVTYPE_NTSC)
        {
            write_integrated_tv_patch_table(ITV_PATCH_TABLE_RGB_NTSC);
        }
        else if (tv_setting_info.system == TVTYPE_PAL)
        {
            write_integrated_tv_patch_table(ITV_PATCH_TABLE_RGB_PAL);
        }
    }
    else if (tv_setting_info.out_signal == TV_OUTPUT_YPBPR)
    {
        if (tv_setting_info.system == TVTYPE_NTSC)
        {
            write_integrated_tv_patch_table(ITV_PATCH_TABLE_YPBPR_NTSC);
        }
        else if (tv_setting_info.system == TVTYPE_PAL)
        {
            write_integrated_tv_patch_table(ITV_PATCH_TABLE_YPBPR_PAL);
        }
    }
}

void patch_for_macrovision(void)
{ 
    /* We should do this patch for this issue:  */
    /* The H-sync cycle of VT3324A3 integrated TV disappeared when disable TV macrovision. */
    if (CX700_REVISION_700M2 != chip_info.gfx_chip_revision) /* Macrovision disable. */
    {
        switch (tv_setting_info.system)
        {
            case TVTYPE_NTSC:
            case TVTYPE_480P:
                access_integrated_tv_regs(ITV_REG_ID_PBPR_BLANK_LEVEL, 0x6D, ITV_CMD_WRITE);
                break;
                
            case TVTYPE_PAL:
            case TVTYPE_576P:
                access_integrated_tv_regs(ITV_REG_ID_PBPR_BLANK_LEVEL, 0x71, ITV_CMD_WRITE);
                break;

            case TVTYPE_720P:
            case TVTYPE_1080I:
                access_integrated_tv_regs(ITV_REG_ID_PBPR_BLANK_LEVEL, 0x72, ITV_CMD_WRITE);
                break;
        }        
    }
}


void set_integrated_tv_mode(int tv_mode_index)
{
    int i, mode_index;
    unsigned long  *tv_mode_table = NULL;
    unsigned long  *tv_crtc_table = NULL;

    DEBUG_MSG(KERN_INFO "set_integrated_tv_mode!\n");
    DEBUG_MSG(KERN_INFO "TV Mode = %d.\n", tv_mode_index);
    DEBUG_MSG(KERN_INFO "TV Type = %d\n", tv_setting_info.system);
    DEBUG_MSG(KERN_INFO "TV Output = %d\n", tv_setting_info.out_signal);
    DEBUG_MSG(KERN_INFO "TV IGA = %d\n", tv_setting_info.iga_path);    
    
    /* Try to get the suitable mode table. */
    for (i=0; i<NUM_TOTAL_INTEGRATEDTV_MODE_TABLE; i++)
    {
        mode_index = i;

        if (ITV_MODE_TABLE[i].mode_index == tv_mode_index)
        {
            break;
        }
    }
    
    DEBUG_MSG(KERN_INFO "TV Mode Index to use is: %d.\n" , mode_index);

    switch (tv_setting_info.system)
    {
        case TVTYPE_NTSC:
        {
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_ntsc;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_ntsc;
            break;
        }
            
        case TVTYPE_PAL:
        {
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_pal;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_pal;
            break;
        }

        case TVTYPE_480P:
        {
            /* For 480P, we use NTSC's mode table, and then do some patches later. */
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_ntsc;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_ntsc;
            break;
        }

        case TVTYPE_576P:
        {
            /* For 576P, we use PAL's mode table, and then do some patches later. */
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_pal;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_pal;
            break;
        }

        case TVTYPE_720P:
        {
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_720p;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_720p;
            break;
        }

        case TVTYPE_1080I:
        {
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_1080i;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_1080i;
            break;
        }

        default:
        {
            tv_mode_table = ITV_MODE_TABLE[mode_index].std_ntsc;
            tv_crtc_table = ITV_CRTC_TABLE[mode_index].std_ntsc;
            break;
        }
    }
   
    enable_integrated_tv_shadow_reg();

    /* Write TV Encoder Regsiters: */

    /* Common settings by mode: */
    write_integrated_tv_general_table(ITV_MODE_TABLE_OFFSET, tv_mode_table);
    
    /* Write GFX CRTC timing by TV mode. */
    write_gfx_crtc_timing_for_tv(tv_crtc_table);

    if (tv_setting_info.iga_path == IGA2)
    {
        access_integrated_tv_regs(ITV_REG_ID_SRC_SEL, ITV_VALUE_SRC_FROM_IGA2, ITV_CMD_WRITE);
        
        write_reg_mask(CR6A, VIACR, 0, BIT2);

        enable_second_display_channel();
    }
    else
    {
        access_integrated_tv_regs(ITV_REG_ID_SRC_SEL, ITV_VALUE_SRC_FROM_IGA1, ITV_CMD_WRITE);
        
        /* For the case: TV IGA1 horizontal total can't be divided by 8. */
        plus_vck_to_iga1_timing(tv_crtc_table[H_TOTAL_INDEX]);
    }

    /* Set TV output signal: */
    set_integrated_tv_output_signal();

    /* Patch for HDTV: */
    set_hdtv_patch_table(mode_index);

    /* We should modify some settings if use RGB/YPbPr output. */
    set_tv_output_patch_table();

    /* Update scaling factor: */
    update_scaling_factor();

    /* Ported from X to support A3*/
    patch_for_macrovision();

    disable_integrated_tv_shadow_reg();
}


void integrated_tv_sw_reset(void)
{
    enable_integrated_tv_shadow_reg();

    access_integrated_tv_regs(ITV_REG_ID_SD_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_SW_RESET, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_SW_RESET, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    delays(50000);  /* We should delay to wait IGA1's VSync. */
    
    access_integrated_tv_regs(ITV_REG_ID_SD_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);

    disable_integrated_tv_shadow_reg();
}


void integrated_tv_enable(void)
{
    /* Turn on DPMS. */
    write_reg_mask(CR36, VIACR, 0, BIT4+BIT5);

    enable_integrated_tv_shadow_reg();

    /* For the problem TV may crash when we repeat to switch TV type */
    /* (NTSC->PAL->NTSC->PAL->...), to do the following sequence can help avoiding this problem. */
    /* 1. Disable Sync Buffer Mode before set mode and disable TV. */
    /* 2. Enable Sync Buffer Mode before enable TV. */    
    /* Suggested by HW team. */

    /* Enable Sync Buffer Mode: */
    access_integrated_tv_regs(ITV_REG_ID_SYNC_BUF_MODE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
   
    /* Enable PK. */
    access_integrated_tv_regs(ITV_REG_ID_PK_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);

    /* Enable SD. */
    access_integrated_tv_regs(ITV_REG_ID_SD_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);

    /* Enable TV encoder. */
    access_integrated_tv_regs(ITV_REG_ID_TV_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);

    /* Change DAC usage. */
    access_integrated_tv_regs(ITV_REG_ID_PRI_DAC_SEL, ITV_VALUE_DAC_FOR_TV, ITV_CMD_WRITE);

    /* Disable SW Reset. */
    access_integrated_tv_regs(ITV_REG_ID_SW_RESET, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    disable_integrated_tv_shadow_reg();

    integrated_tv_sw_reset();
}


void integrated_tv_disable(void)
{
    enable_integrated_tv_shadow_reg();

    /* For the problem TV may crash when we repeat to switch TV type */
    /* (NTSC->PAL->NTSC->PAL->...), to do the following sequence can help avoiding this problem. */
    /* 1. Disable Sync Buffer Mode before set mode and disable TV. */
    /* 2. Enable Sync Buffer Mode before enable TV. */    
    /* Suggested by HW team. */

    /* Enable SW Reset. */
    access_integrated_tv_regs(ITV_REG_ID_SW_RESET, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
    
    /* Disable Sync Buffer Mode: */
    access_integrated_tv_regs(ITV_REG_ID_SYNC_BUF_MODE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    /* Disable PK. */
    access_integrated_tv_regs(ITV_REG_ID_PK_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    /* Disable SD. */
    access_integrated_tv_regs(ITV_REG_ID_SD_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    /* Disable TV encoder. */
    access_integrated_tv_regs(ITV_REG_ID_TV_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);

    /* Change DAC usage. */
    access_integrated_tv_regs(ITV_REG_ID_PRI_DAC_SEL, ITV_VALUE_DAC_FOR_CRT, ITV_CMD_WRITE);

    disable_integrated_tv_shadow_reg();
}


void get_integrated_tv_default_status(void)
{
    static int scale_value_initialized;
    enable_integrated_tv_shadow_reg();

    /* Size: */
    /* The scaling factor level and default value here are not absolute, they can change by request. */
    tv_setting_info.ScalHLevel = 21;
    tv_setting_info.ScalVLevel = 15;
    tv_setting_info.DefaultScalH = 10;
    tv_setting_info.DefaultScalV = 7;
    if(!scale_value_initialized) {
        tv_setting_info.CurrentScalH = tv_setting_info.DefaultScalH;
        tv_setting_info.CurrentScalV = tv_setting_info.DefaultScalV;
        scale_value_initialized=1;
    }
    tv_setting_info.ScalDegreeH = 5;
    tv_setting_info.ScalDegreeV = 5;

    /* Positoin: */
    /* The position level and default value here are not absolute, they can change by request. */
    tv_setting_info.PositionHLevel = 17;
    tv_setting_info.PositionVLevel = 13;
    tv_setting_info.DefaultPositionH = 9;
    tv_setting_info.DefaultPositionV = 7;
    tv_setting_info.CurrentPositionH = tv_setting_info.DefaultPositionH;
    tv_setting_info.CurrentPositionV = tv_setting_info.DefaultPositionV;

    tv_setting_info.DefaultFFilter = 0;
    tv_setting_info.FFilterLevel = 1;
    tv_setting_info.CurrentFFilter = tv_setting_info.DefaultFFilter;

    tv_setting_info.DefaultAFFilter = 0;
    tv_setting_info.AFFilterLevel = 1023;
    
    /* Brightness: */
    tv_setting_info.DefaultBrightness = access_integrated_tv_regs(ITV_REG_ID_BLACK_LEVEL, 0, ITV_CMD_READ);
    tv_setting_info.BrightnessLevel = 255;  /* 8 bits. */
    tv_setting_info.CurrentBrightness = tv_setting_info.DefaultBrightness;

    /* Contrast: */
    tv_setting_info.DefaultContrast= access_integrated_tv_regs(ITV_REG_ID_CONTRAST, 0, ITV_CMD_READ);
    tv_setting_info.ContrastLevel = 512;  /* 9 bits. */
    tv_setting_info.CurrentContrast= tv_setting_info.DefaultContrast;

    /* Saturation: */
    tv_setting_info.DefaultSaturation= access_integrated_tv_regs(ITV_REG_ID_SATURATION, 0, ITV_CMD_READ);
    tv_setting_info.SaturationLevel = 0x3FFFF;  /* 9+9 bits. */
    tv_setting_info.CurrentSaturation= tv_setting_info.DefaultSaturation;

    /* Hue: */
    tv_setting_info.DefaultTINT = access_integrated_tv_regs(ITV_REG_ID_HUE, 0, ITV_CMD_READ);
    tv_setting_info.TINTLevel = 2047;       /* 11 bits. */
    tv_setting_info.CurrentTINT = tv_setting_info.DefaultTINT;
    
    disable_integrated_tv_shadow_reg();
}


int get_integrated_tv_ffilter_status(int filter_type, int* filter_state)
{
    int value = 0;
    int tmp_filter_state = 0;
    int tmp_filter_type =0;
    
    enable_integrated_tv_shadow_reg();

    /* check if filter is enable */
    tmp_filter_state = access_integrated_tv_regs(ITV_REG_ID_DE_FLICKER_ENABLE, 0, ITV_CMD_READ);
 
    /* get current filter type */
    tmp_filter_type = access_integrated_tv_regs(ITV_REG_ID_DE_FLK_NUM, 0, ITV_CMD_READ);
    
    if (filter_type == ITV_FILTER_TYPE_NORMAL)
    {
        if ((tmp_filter_state == STATE_ON) && (tmp_filter_type == ITV_VALUE_FLK_NUM_WHOLE_IMAGE))
        {
            *filter_state = STATE_ON;
        }
        else
        {
            *filter_state = STATE_OFF;
        }
        value = access_integrated_tv_regs(ITV_REG_ID_DF_WT_SEL, 0, ITV_CMD_READ);
    }
    else if (filter_type == ITV_FILTER_TYPE_ADAPTIVE)
    {
        if ((tmp_filter_state == STATE_ON) && (tmp_filter_type == ITV_VALUE_FLK_NUM_ONE_POINT))
        {
            *filter_state = STATE_ON;
        }
        else
        {
            *filter_state = STATE_OFF;
        }
        value = access_integrated_tv_regs(ITV_REG_ID_DE_FLK_THRESHOLD, 0, ITV_CMD_READ);
    }
    
    disable_integrated_tv_shadow_reg();

    return value;
}



void get_integrated_tv_pos(u32* current_hpos, u32* current_vpos)
{        
    enable_integrated_tv_shadow_reg();

    *current_hpos = access_integrated_tv_regs(ITV_REG_ID_DST_HSD, 0, ITV_CMD_READ);
    *current_vpos = access_integrated_tv_regs(ITV_REG_ID_DST_VSD, 0, ITV_CMD_READ);

    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_ffilter_status(int filter_type, int filter_on_off, u32 desired_value)
{
    enable_integrated_tv_shadow_reg();

    if (filter_type == ITV_FILTER_TYPE_NORMAL)
    {
        if (filter_on_off == ITV_FILTER_ON)
        {
            /* Turn on FFilter: */
            access_integrated_tv_regs(ITV_REG_ID_DE_FLICKER_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
            access_integrated_tv_regs(ITV_REG_ID_DF_TYPE, ITV_FILTER_TYPE_NORMAL, ITV_CMD_WRITE);

            /* Set the weighting of FFilter : */
            if (desired_value < 0)
            {
                desired_value = 0;
            }

            if (desired_value > tv_setting_info.FFilterLevel)
            {
                desired_value = tv_setting_info.FFilterLevel;
            }

            access_integrated_tv_regs(ITV_REG_ID_DF_WT_SEL, desired_value, ITV_CMD_WRITE);
            
            tv_setting_info.CurrentFFilter = desired_value;
        }
        else
        {
            /* Turn off FFilter: */
            access_integrated_tv_regs(ITV_REG_ID_DE_FLICKER_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);
        }
    }
    else if (filter_type == ITV_FILTER_TYPE_ADAPTIVE)
    {
        if (filter_on_off == ITV_FILTER_ON)
        {
            /* Turn on AFFilter: */
            access_integrated_tv_regs(ITV_REG_ID_DE_FLICKER_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
            access_integrated_tv_regs(ITV_REG_ID_DE_FLK_NUM, ITV_VALUE_FLK_NUM_ONE_POINT, ITV_CMD_WRITE);

            /* Set threshold of AFFilter : */
            if (desired_value < 0)
            {
                desired_value = 0;
            }
            
            if (desired_value > tv_setting_info.AFFilterLevel)
            {
                desired_value = tv_setting_info.AFFilterLevel;
            }

            access_integrated_tv_regs(ITV_REG_ID_DE_FLK_THRESHOLD, desired_value, ITV_CMD_WRITE);
        }
        else
        {
            /* Turn off AFFilter: */
            access_integrated_tv_regs(ITV_REG_ID_DE_FLK_NUM, ITV_VALUE_FLK_NUM_WHOLE_IMAGE, ITV_CMD_WRITE);
            access_integrated_tv_regs(ITV_REG_ID_DE_FLK_THRESHOLD, 0, ITV_CMD_WRITE);
            access_integrated_tv_regs(ITV_REG_ID_DE_FLICKER_ENABLE, ITV_VALUE_DISABLE, ITV_CMD_WRITE);
        }
    }

    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_hue(u32 desired_value)
{
    enable_integrated_tv_shadow_reg();
    access_integrated_tv_regs(ITV_REG_ID_HUE, desired_value, ITV_CMD_WRITE);
    tv_setting_info.CurrentTINT = desired_value;
    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_brightness(u32 desired_value)
{
    enable_integrated_tv_shadow_reg();
    access_integrated_tv_regs(ITV_REG_ID_BLACK_LEVEL, desired_value, ITV_CMD_WRITE);
    tv_setting_info.CurrentBrightness = desired_value;
    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_contrast(u32 desired_value)
{
    enable_integrated_tv_shadow_reg();
    access_integrated_tv_regs(ITV_REG_ID_CONTRAST, desired_value, ITV_CMD_WRITE);
    tv_setting_info.CurrentContrast = desired_value;
    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_saturation(u32 desired_value)
{
    enable_integrated_tv_shadow_reg();
    access_integrated_tv_regs(ITV_REG_ID_SATURATION, desired_value, ITV_CMD_WRITE);
    tv_setting_info.CurrentSaturation = desired_value;
    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_pos(u32 desired_hpos, u32 desired_vpos)
{
    u32 dst_hpos, dst_vpos;
    
    enable_integrated_tv_shadow_reg();

    dst_hpos = access_integrated_tv_regs(ITV_REG_ID_DST_HSD, 0, ITV_CMD_READ);
    dst_vpos = access_integrated_tv_regs(ITV_REG_ID_DST_VSD, 0, ITV_CMD_READ);

    dst_hpos = dst_hpos + (desired_hpos - tv_setting_info.CurrentPositionH);
    dst_vpos = dst_vpos - (desired_vpos - tv_setting_info.CurrentPositionV);

    access_integrated_tv_regs(ITV_REG_ID_DST_HSD, dst_hpos, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_DST_VSD, dst_vpos, ITV_CMD_WRITE);

    tv_setting_info.CurrentPositionH = desired_hpos;
    tv_setting_info.CurrentPositionV = desired_vpos;

    disable_integrated_tv_shadow_reg();
}


void set_integrated_tv_size(u32 desired_hsize, u32 desired_vsize)
{ 
    u32 dst_hsize, dst_vsize;
    DEBUG_MSG("set integrated TV H:%d V:%d", desired_hsize, desired_vsize);
    
    enable_integrated_tv_shadow_reg();
    
    dst_hsize = access_integrated_tv_regs(ITV_REG_ID_DST_HSIZE, 0, ITV_CMD_READ);
    dst_vsize = access_integrated_tv_regs(ITV_REG_ID_DST_VSIZE, 0, ITV_CMD_READ);

    dst_hsize = dst_hsize + tv_setting_info.ScalDegreeH * (desired_hsize - tv_setting_info.CurrentScalH);
    dst_vsize = dst_vsize + tv_setting_info.ScalDegreeV * (desired_vsize - tv_setting_info.CurrentScalV);

    access_integrated_tv_regs(ITV_REG_ID_DST_HSIZE, dst_hsize, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_DST_VSIZE, dst_vsize, ITV_CMD_WRITE);

    /* Update scaling factor: */
    update_scaling_factor();
    
    tv_setting_info.CurrentScalH = desired_hsize;
    tv_setting_info.CurrentScalV = desired_vsize;
    disable_integrated_tv_shadow_reg();
}


int integrated_tv_sense(void)
{
    u8 cr30, cr37, cr43, rgb_dac_status;
    u32 tv3f, rgb_status, tv_encoder_status, dac_status;
    int sense_status;

    enable_integrated_tv_shadow_reg();

    if (g_first_time_to_sense)
    {
        /* Set TV mode (arbitrary) at least once before sense TV. */
        write_integrated_tv_general_table(ITV_INIT_TABLE_OFFSET, ITV_INIT_TABLE_VALUE);
        write_integrated_tv_general_table(ITV_MODE_TABLE_OFFSET, ITV_MODE_TABLE_NTSC_640x480);
        g_first_time_to_sense = FALSE;
    }

    /* Saved for restore later: */
    cr30 = read_reg(VIACR, CR30);
    cr37 = read_reg(VIACR, CR37);
    cr43 = read_reg(VIACR, CR43);
    rgb_dac_status = read_reg(VIASR, SR5E);
    tv3f = access_integrated_tv_regs(ITV_REG_ID_TV_SENSE_DATA, 0, ITV_CMD_READ);
    tv_encoder_status = access_integrated_tv_regs(ITV_REG_ID_TV_ENABLE, 0, ITV_CMD_READ);
    dac_status = access_integrated_tv_regs(ITV_REG_ID_PRI_DAC_SEL, 0, ITV_CMD_READ);

    /* To sense TV, we should: */
    /* 1. Set CR30[4:3]=01, CR37[2:0]=100, CR37[3]=0, CR43[5:4]=01 and TV.3F[7:0]=42h. */
    /* 2. Enable TV DAC R/G/B. */
    /* 3. Enable TV encoder. */
    /* 4. Change primary DAC to TV. */
    write_reg_mask(CR30, VIACR, BIT3, BIT4+BIT3);
    write_reg_mask(CR37, VIACR, BIT2, BIT3+BIT2+BIT1+BIT0);
    write_reg_mask(CR43, VIACR, BIT4, BIT5+BIT4);
    write_reg_mask(SR5E, VIASR, 0, BIT3+BIT2+BIT1);
    access_integrated_tv_regs(ITV_REG_ID_TV_SENSE_DATA, 0x42, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_TV_ENABLE, ITV_VALUE_ENABLE, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_PRI_DAC_SEL, ITV_VALUE_DAC_FOR_TV, ITV_CMD_WRITE);
    
    /* Start to sense: */
    access_integrated_tv_regs(ITV_REG_ID_TV_SENSE, ITV_VALUE_MANUAL_SENSE, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_TV_SENSE, ITV_VALUE_NO_SENSE, ITV_CMD_WRITE);

    /* Restore: */
    access_integrated_tv_regs(ITV_REG_ID_PRI_DAC_SEL, dac_status, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_TV_ENABLE, tv_encoder_status, ITV_CMD_WRITE);
    access_integrated_tv_regs(ITV_REG_ID_TV_SENSE_DATA, tv3f, ITV_CMD_WRITE);
    write_reg(CR30, VIACR, cr30);
    write_reg(CR37, VIACR, cr37);
    write_reg(CR43, VIACR, cr43);
    write_reg(SR5E, VIASR, rgb_dac_status);

    /* Get sense result: */    
    sense_status = access_integrated_tv_regs(ITV_REG_ID_TV_SENSE_STATUS, 0, ITV_CMD_READ);
    rgb_status = access_integrated_tv_regs(ITV_REG_ID_TV_SENSE_RGB_STATUS, 0, ITV_CMD_READ);

    /* Check connected TV output: */
    if (!tv_setting_info.out_signal)
    {
        switch(rgb_status)
        {
            case ITV_VALUE_DAC_R | ITV_VALUE_DAC_G:
            {
                tv_setting_info.out_signal= TV_OUTPUT_SVIDEO;
                break;
            }

            case ITV_VALUE_DAC_B:
            {
                tv_setting_info.out_signal = TV_OUTPUT_COMPOSITE;
                break;
            }

            case ITV_VALUE_DAC_R | ITV_VALUE_DAC_G | ITV_VALUE_DAC_B:
            {
                tv_setting_info.out_signal = TV_OUTPUT_RGB;
                break;
            }

            default:
            {
                tv_setting_info.out_signal = TV_OUTPUT_COMPOSITE;
                break;
            }
        }
    }

    disable_integrated_tv_shadow_reg();

    return sense_status;
}


int init_integrated_tv_regs(void)
{
    u32 mask = 0x1FFFFFF0;  /* [28:4] */
    u32 buf_start_address;
    u32 buffer_size = 0x80000;
    u8  offset = 0;
    unsigned char* mmio_base;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    buf_start_address = viafbinfo.memsize - INTEGRATED_TV_BUFFER_SIZE;
    
    viafbinfo.integrated_tv_map_base = (unsigned int)ioremap_nocache(viafbinfo.mmio_base  + VIA_MMIO_INTEGRATED_TV_BASE, VIA_MMIO_INTEGRATED_TV_SIZE);

    if (!viafbinfo.integrated_tv_map_base)
    {        
        return FAIL;
    }

    mmio_base = (unsigned char*)viafbinfo.integrated_tv_map_base;
#else
    buf_start_address = parinfo.fbmem_free - INTEGRATED_TV_BUFFER_SIZE;
    parinfo.fbmem_free = parinfo.fbmem_free - INTEGRATED_TV_BUFFER_SIZE;
    parinfo.fbmem_used += INTEGRATED_TV_BUFFER_SIZE;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
    parinfo.integrated_tv_map_base = ioremap_nocache(parinfo.mmio_base + VIA_MMIO_INTEGRATED_TV_BASE, VIA_MMIO_INTEGRATED_TV_SIZE);
#else
    parinfo.integrated_tv_map_base = (unsigned int)ioremap_nocache(parinfo.mmio_base + VIA_MMIO_INTEGRATED_TV_BASE, VIA_MMIO_INTEGRATED_TV_SIZE);
#endif

    if (!parinfo.integrated_tv_map_base)
    {        
        return FAIL;
    }

    mmio_base = (unsigned char*)parinfo.integrated_tv_map_base;
#endif

    enable_integrated_tv_shadow_reg();

    /* Clear all: */
    while (offset < 0xF0)
    {
        write_integrated_tv_reg32(mmio_base, offset, 0);
        offset = offset + 4;
    }
    
    /* Write Source Buffer 0~3 Base Address: */
    write_integrated_tv_reg_mask32(mmio_base, 0x04, mask, buf_start_address);

    buf_start_address = buf_start_address + buffer_size;
    write_integrated_tv_reg_mask32(mmio_base, 0x08, mask, buf_start_address);

    buf_start_address = buf_start_address + buffer_size;
    write_integrated_tv_reg_mask32(mmio_base, 0x0C, mask, buf_start_address);

    buf_start_address = buf_start_address + buffer_size;
    write_integrated_tv_reg_mask32(mmio_base, 0x10, mask, buf_start_address);
    
    /* Write Destination Even and Odd Buffer Base Address: */
    buf_start_address = buf_start_address + buffer_size;
    write_integrated_tv_reg_mask32(mmio_base, 0x2C, mask, buf_start_address);

    buf_start_address = buf_start_address + buffer_size;
    write_integrated_tv_reg_mask32(mmio_base, 0x30, mask, buf_start_address);

    /* Other initialize: */
    write_integrated_tv_general_table(ITV_INIT_TABLE_OFFSET, ITV_INIT_TABLE_VALUE);

    disable_integrated_tv_shadow_reg();

    return OK;
}


int integrated_tv_encoder_identify(void)
{
    if (chip_info.gfx_chip_name == UNICHROME_CX700)
    {
        return OK;
    }
    else
    {
        return FAIL;
    }
}

