 /*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 

#include <linux/fb.h>
#include <asm/uaccess.h>

#include "viafbdev.h"
#include "chip.h"
#include "debug.h"
#include "accel.h"
#include "share.h"

/* External struct */

extern struct chip_information      chip_info;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)

void Init_Accel(void);
void Init_2D_Engine(void);
void Set_2D_Color_Depth(int);
void HWCursorInit(void);
void ShowHideHWCursor(struct fb_info *info, int Status);
int WaitEngineIdle(void);

extern struct viafb_par parinfo;


void Init_Accel(void)
{
    parinfo.cursor_start = parinfo.fbmem_free - CURSOR_SIZE;
    parinfo.fbmem_free -= CURSOR_SIZE;
    parinfo.fbmem_used += CURSOR_SIZE;

    /* Reverse 8*1024 memory space for cursor image of the XServer */
    parinfo.VQ_start = parinfo.fbmem_free - CURSOR_SIZE - VQ_SIZE;
    parinfo.VQ_end = parinfo.VQ_start + VQ_SIZE - 1 ;
    parinfo.fbmem_free -= (CURSOR_SIZE + VQ_SIZE);
    parinfo.fbmem_used += (CURSOR_SIZE + VQ_SIZE);
}

void Init_2D_Engine(void)
{
    u32 dwVQStartAddr, dwVQEndAddr;
    u32 dwVQLen, dwVQStartL, dwVQEndL, dwVQStartEndH;

    /* init 2D engine regs to reset 2D engine */
    MMIO_OUT32(VIA_REG_GEMODE,     0x0);
    MMIO_OUT32(VIA_REG_SRCPOS,     0x0);
    MMIO_OUT32(VIA_REG_DSTPOS,     0x0);
    MMIO_OUT32(VIA_REG_DIMENSION,  0x0);
    MMIO_OUT32(VIA_REG_PATADDR,    0x0);
    MMIO_OUT32(VIA_REG_FGCOLOR,    0x0);
    MMIO_OUT32(VIA_REG_BGCOLOR,    0x0);
    MMIO_OUT32(VIA_REG_CLIPTL,     0x0);
    MMIO_OUT32(VIA_REG_CLIPBR,     0x0);
    MMIO_OUT32(VIA_REG_OFFSET,     0x0);
    MMIO_OUT32(VIA_REG_KEYCONTROL, 0x0);
    MMIO_OUT32(VIA_REG_SRCBASE,    0x0);
    MMIO_OUT32(VIA_REG_DSTBASE,    0x0);
    MMIO_OUT32(VIA_REG_PITCH,      0x0);
    MMIO_OUT32(VIA_REG_MONOPAT1,   0x0);

    /* Init AGP and VQ regs */
    switch(chip_info.gfx_chip_name)
    {
        /* Modified from X driver code to support VT3336 */
        case UNICHROME_K8M890:
            MMIO_OUT32(0x41c, 0x00100000);
            MMIO_OUT32(0x420, 0x680A0000);
            MMIO_OUT32(0x420, 0x02000000);
            break; 
            
        default:
            MMIO_OUT32(VIA_REG_TRANSET,   0x00100000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x00000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x00333004);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x60000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x61000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x62000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x63000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x64000000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x7D000000);

            MMIO_OUT32(VIA_REG_TRANSET,   0xFE020000);
            MMIO_OUT32(VIA_REG_TRANSPACE, 0x00000000);
            break;
    }
    if (parinfo.VQ_start != 0) {
        /* Enable VQ */
        dwVQStartAddr = parinfo.VQ_start;
        dwVQEndAddr = parinfo.VQ_end;
        switch(chip_info.gfx_chip_name)
        {
            /* Modified from X driver code to support VT3336 */
            case UNICHROME_K8M890:
                dwVQStartL = 0x70000000 | (dwVQStartAddr & 0xFFFFFF);
                dwVQEndL = 0x71000000 | (dwVQEndAddr & 0xFFFFFF);
                dwVQStartEndH = 0x72000000 | ((dwVQStartAddr & 0xFF000000) >> 24) |
                                ((dwVQEndAddr & 0xFF000000) >> 16);
                dwVQLen = 0x73000000 | (VQ_SIZE >> 3);
                break;
            default:
                dwVQStartL = 0x50000000 | (dwVQStartAddr & 0xFFFFFF);
                dwVQEndL = 0x51000000 | (dwVQEndAddr & 0xFFFFFF);
                dwVQStartEndH = 0x52000000 | ((dwVQStartAddr & 0xFF000000) >> 24) |
                                ((dwVQEndAddr & 0xFF000000) >> 16);
                dwVQLen = 0x53000000 | (VQ_SIZE >> 3);
                break;
        }

        switch(chip_info.gfx_chip_name)
        {
            /* Modified from X driver code to support VT3336 */
            case UNICHROME_K8M890:
                MMIO_OUT32(0x41c, 0x00100000);
                MMIO_OUT32(0x420, dwVQStartEndH);
                MMIO_OUT32(0x420, dwVQStartL);
                MMIO_OUT32(0x420, dwVQEndL);
                MMIO_OUT32(0x420, dwVQLen);
                MMIO_OUT32(0x420, 0x74301001);
                MMIO_OUT32(0x420, 0x00000000);
                break;
            default:
                MMIO_OUT32(VIA_REG_TRANSET,   0x00FE0000);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x080003FE);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0A00027C);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0B000260);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0C000274);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0D000264);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0E000000);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x0F000020);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x1000027E);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x110002FE);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x200F0060);

                MMIO_OUT32(VIA_REG_TRANSPACE, 0x00000006);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x40008C0F);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x44000000);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x45080C04);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x46800408);

                MMIO_OUT32(VIA_REG_TRANSPACE, dwVQStartEndH);
                MMIO_OUT32(VIA_REG_TRANSPACE, dwVQStartL);
                MMIO_OUT32(VIA_REG_TRANSPACE, dwVQEndL);
                MMIO_OUT32(VIA_REG_TRANSPACE, dwVQLen);
                break;
        }
    }
    else {
        /* Diable VQ */
        switch(chip_info.gfx_chip_name)
        {
            /* Modified from X driver code to support VT3336 VT3327 */
            case UNICHROME_K8M890:
            case UNICHROME_P4M890:
                MMIO_OUT32(0x41c, 0x00100000);
                MMIO_OUT32(0x420, 0x74301000);
                break;
            default:
                MMIO_OUT32(VIA_REG_TRANSET,   0x00FE0000);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x00000004);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x40008C0F);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x44000000);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x45080C04);
                MMIO_OUT32(VIA_REG_TRANSPACE, 0x46800408);
                break;
       }
    }

    Set_2D_Color_Depth(parinfo.bpp);

    MMIO_OUT32(VIA_REG_SRCBASE, 0x0);
    MMIO_OUT32(VIA_REG_DSTBASE, 0x0);

    MMIO_OUT32(VIA_REG_PITCH, VIA_PITCH_ENABLE | (((parinfo.hres * parinfo.bpp >> 3) >> 3) |
                (((parinfo.hres * parinfo.bpp >> 3) >> 3) << 16)));
}

void Set_2D_Color_Depth(int bpp)
{
    u32 dwGEMode;

    dwGEMode = MMIO_IN32(0x04) & 0xFFFFFCFF;

    switch(bpp)
    {
        case 16:
            dwGEMode |= VIA_GEM_16bpp;
            break;
        case 32:
            dwGEMode |= VIA_GEM_32bpp;
            break;
        default:
            dwGEMode |= VIA_GEM_8bpp;
            break;
    }

    /* Set BPP and Pitch */
    MMIO_OUT32(VIA_REG_GEMODE, dwGEMode);
}

void HWCursorInit(void)
{
    /* Set Cursor Image Base Address */
    MMIO_OUT32(VIA_REG_CURSOR_MODE, parinfo.cursor_start);
    MMIO_OUT32(VIA_REG_CURSOR_POS,  0x0);
    MMIO_OUT32(VIA_REG_CURSOR_ORG,  0x0);
    MMIO_OUT32(VIA_REG_CURSOR_BG,   0x0);
    MMIO_OUT32(VIA_REG_CURSOR_FG,   0x0);
}

void ShowHideHWCursor(struct fb_info *info, int Status)
{
    u32 temp;
    u32 iga_path = ((struct viafb_par *)(info->par))->iga_path;

    temp = MMIO_IN32(VIA_REG_CURSOR_MODE);
    switch(Status)
    {
        case HW_Cursor_ON:
            temp |= 0x1;
            break;
        case HW_Cursor_OFF:
            temp &= 0xFFFFFFFE;
            break;
    }
    switch(iga_path)
    {
        case IGA2:
            temp |=0x80000000;
            break;
        case IGA1:
        default:
            temp &=0x7FFFFFFF;
    }
    MMIO_OUT32(VIA_REG_CURSOR_MODE, temp);
}

int WaitEngineIdle(void)
{
    int loop = 0;

    while (!(MMIO_IN32(VIA_REG_STATUS) & VIA_VR_QUEUE_BUSY) && (loop++ < MAXLOOP))
        ;

    while ((MMIO_IN32(VIA_REG_STATUS) &
          (VIA_CMD_RGTR_BUSY | VIA_2D_ENG_BUSY | VIA_3D_ENG_BUSY)) &&
          (loop++ < MAXLOOP))
        ;

    return loop >= MAXLOOP;
}

#endif

