/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef __CHIP_H__
#define __CHIP_H__

#include "tv.h"

/***************************************/
/* Definition Graphic Chip Information */
/***************************************/

#define     PCI_VIA_VENDOR_ID       0x1106

/* Define VIA Graphic Chip Name */
#define     UNICHROME_CLE266        1
#define     UNICHROME_CLE266_DID    0x3122
#define     CLE266_REVISION_AX      0x0A
#define     CLE266_REVISION_CX      0x0C

#define     UNICHROME_K400          2
#define     UNICHROME_K400_DID      0x7205

#define     UNICHROME_K800          3
#define     UNICHROME_K800_DID      0x3108


#define     UNICHROME_P880          4
#define     UNICHROME_P880_DID      0x3118

#define     UNICHROME_CN900         5
#define     UNICHROME_CN900_DID     0x3344

#define     UNICHROME_CX700         6
#define     UNICHROME_CX700_DID     0x3157
#define     CX700_REVISION_700         0x0
#define     CX700_REVISION_700M      0x1
#define     CX700_REVISION_700M2    0x2

#define     UNICHROME_CN750         7
#define     UNICHROME_CN750_DID     0x3225

#define     UNICHROME_K8M890       8
#define     UNICHROME_K8M890_DID  0x3230

#define     UNICHROME_P4M890      9
#define     UNICHROME_P4M890_DID  0x3343

#define     UNICHROME_P4M900      10
#define     UNICHROME_P4M900_DID  0x3371





/*************************************/
/* Definition TV Encoder Information */
/*************************************/

/* Definition TV Encoder Index */

#define     NON_TV_ENCODER      0x00
#define     VT1621              0x01
#define     VT1622              0x02
#define     CH7009              0x03
#define     CH7019              0x04
#define     SAA7108A            0x05
#define     CH7005              0x06
#define     VT1622A             0x07
#define     VT1623              0x08
#define     FS453               0x09
#define     FS454               0x0A
#define     VT1625              0x0B
#define     INTEGRATED_TV       0x0C

/* Definition TV Encoder I2C Slave Address */
#define     VT1622_TV_I2C_ADDR      0x40
#define     VT1622A_TV_I2C_ADDR     0x40
#define     VT1625_TV_I2C_ADDR      0x40


#define     TV_ON_AMR       1
#define     TV_ON_AGP       2

/**************************************************/
/* Definition TMDS Trasmitter Information         */
/**************************************************/

/* Definition TMDS Trasmitter Index */
#define     NON_TMDS_TRANSMITTER    0x00
#define     VT1632_TMDS             0x01
#define     SII164_TMDS             0x02
#define     INTEGRATED_TMDS         0x42

/* Definition TMDS Trasmitter I2C Slave Address */
#define     VT1632_TMDS_I2C_ADDR    0x10
#define     SII164_TMDS_I2C_ADDR    0x70

/**************************************************/
/* Definition LVDS Trasmitter Information         */
/**************************************************/

/* Definition LVDS Trasmitter Index */
#define     NON_LVDS_TRANSMITTER    0x00
#define     VT1631_LVDS             0x01
#define     VT1636_LVDS             0x0E
#define     INTEGRATED_LVDS         0x41


/* Definition Digital Transmitter Mode */
#define     TX_DATA_12_BITS         0x01
#define     TX_DATA_24_BITS         0x02
#define     TX_DATA_DDR_MODE        0x04
#define     TX_DATA_SDR_MODE        0x08


/* Definition LVDS Trasmitter I2C Slave Address */
#define     VT1631_LVDS_I2C_ADDR    0x70
#define     VT3271_LVDS_I2C_ADDR    0x80
#define     VT1636_LVDS_I2C_ADDR    0x80


/* GFX TV I/O Address */
#define     GFX_TV_IO_ADDR          0x31

struct tv_chip_information {
    int tv_chip_name;
    int tv_chip_slave_addr;
    int tv_chip_on_slot;
    int output_interface;
};

struct tmds_chip_information {
    int tmds_chip_name;
    int tmds_chip_slave_addr;
    int dvi_panel_id;
    int data_mode;
    int output_interface;
    int i2c_port;
};

struct lvds_chip_information {
    int lvds_chip_name;
    int lvds_chip_slave_addr;
    int data_mode;
    int output_interface;
    int i2c_port;
};

struct chip_information  {
    int gfx_chip_name;
    int gfx_chip_revision;
    struct tv_chip_information tv_chip_info;
    struct tmds_chip_information tmds_chip_info;
    struct lvds_chip_information lvds_chip_info;
    struct lvds_chip_information lvds_chip_info2;
};

struct crt_setting_information {
    int iga_path;
    int h_active;
    int v_active;
    int bpp;
    int refresh_rate;
};

struct tv_setting_information {
    int iga_path;
    int h_active;
    int v_active;
    int bpp;
    int system;
    int out_signal;
    int level;
    int dedotcrawl;

    u32  DefaultFFilter;
    u32  CurrentFFilter;
    u32  FFilterLevel;

    u32 DefaultBrightness;
    u32 CurrentBrightness;
    u32 BrightnessLevel;

    u32 DefaultContrast;
    u32 CurrentContrast;
    u32 ContrastLevel;

    u32 DefaultSaturation;
    u32 CurrentSaturation;
    u32 SaturationLevel;

    u32 DefaultTINT;
    u32 CurrentTINT;
    u32 TINTLevel;
    u32 PositionHLevel;
    u32 DefaultPositionH;
    u32 PositionVLevel;
    u32 DefaultPositionV;
    u32 CurrentPositionH;
    u32 CurrentPositionV;
    u32 ScalHLevel;
    u32 ScalVLevel;	
    u32 DefaultScalH;
    u32 DefaultScalV;
    u32 CurrentScalH;
    u32 CurrentScalV;	 
    u32 ScalDegreeH;
    u32 ScalDegreeV;
    u32 AdaptiveFFilterOn;	 
    u32 AFFilterLevel;
    u32 DefaultAFFilter;	
};

struct tmds_setting_information {
    int iga_path;
    int h_active;
    int v_active;
    int bpp;
    int refresh_rate;
    int get_dvi_size_method;
};

struct lvds_setting_information {
    int iga_path;
    int h_active;
    int v_active;
    int bpp;
    int refresh_rate;
    int get_lcd_size_method;
    int lcd_panel_id;
    int lcd_panel_size;
    int lcd_panel_hres;
    int lcd_panel_vres;
    int display_method;
    int LCDDualEdge;
    int LCDDithering;
    int lcd_mode; /* for VT3324 LVDS: SPWG or OPEN_LDI mode */
    u32 vclk; /* panel mode clock value */
};

typedef struct GFX_DPA_SETTING {
    int     ClkRangeIndex;
    u8   DVP0;               /* CR96[3:0] */
    u8   DVP0DataDri_S1;     /* SR2A[5]   */
    u8   DVP0DataDri_S;      /* SR1B[1]   */
    u8   DVP0ClockDri_S1;    /* SR2A[4]   */
    u8   DVP0ClockDri_S;     /* SR1E[2]   */
    u8   DVP1;               /* CR9B[3:0] */
    u8   DVP1Driving;        /* SR65[3:0], Data and Clock driving */
    u8   DFPHigh;            /* CR97[3:0] */
    u8   DFPLow;             /* CR99[3:0] */

} GFXDPASETTING, *GFXDPASETTINGPTR;


typedef struct VT1636_DPA_SETTING {
    int     PanelSizeID;
    u8   CLK_SEL_ST1;
    u8   CLK_SEL_ST2;
} VT1636DPASETTING, *VT1636DPASETTINGPTR;
#endif /* __CHIP_H__ */

