/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include "debug.h"
#include "dvi.h"
#include "chip.h"
#include "share.h"
#include "via_i2c.h"

/* extern function */
extern int i2cWriteByte(u8 slave_addr, u8 index, u8 data);
extern int i2cReadByte(u8 slave_addr, u8 index, u8 *pData);
extern void set_output_path(int device, int set_iga, int output_interface);
extern void fill_crtc_timing(struct crt_mode_table *crt_table, int mode_index, int bpp_byte, int set_iga);
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern inline u8 read_reg(int io_port, u8 index);
extern inline void write_reg(u8 index, u16 io_port, u8 data);

extern struct chip_information      chip_info;
extern struct tmds_setting_information  tmds_setting_info;

extern int EPIA_DVI;
extern int LCDDualEdge;
extern int BusWidth;

int tmds_register_read(int index);
void tmds_register_write(int index, u8 data);
int dvi_sense(void);
int dvi_query_EDID(void);
void dvi_disable(void);
void dvi_enable(void);

extern int display_hardware_layout;

int check_tmds_chip(int device_id_subaddr, int device_id)
{
    if (tmds_register_read(device_id_subaddr) == device_id)
        return(OK);
    else
        return(FAIL);
}

void init_dvi_size(void)
{
    switch(tmds_setting_info.get_dvi_size_method) {
        case GET_DVI_SIZE_BY_SYSTEM_BIOS:
            break;
        case GET_DVI_SZIE_BY_HW_STRAPPING:
            break;
        case GET_DVI_SIZE_BY_VGA_BIOS:
            break;
    }

}

int tmds_trasmitter_identify(void)
{
    unsigned char   sr2a = 0, sr1e = 0, sr3e = 0;
    
    /* Check for VT3324 integrated TMDS. */
    chip_info.tmds_chip_info.tmds_chip_name = INTEGRATED_TMDS;

    /* Not always support intergrated TMDS now,
       we should check support by hardware layout. */
    if ((chip_info.gfx_chip_name == UNICHROME_CX700) &&
        ((display_hardware_layout == HW_LAYOUT_DVI_ONLY) ||
         (display_hardware_layout == HW_LAYOUT_LCD_DVI)))
    {
        DEBUG_MSG(KERN_INFO "\n Integrated TMDS ! \n");
        return(OK);
    }

    /* Turn on ouputting pad */
    /* Modified form X driver's code to support VT3327 VT3336 VT3364 */
    switch (chip_info.gfx_chip_name)
    {
        case UNICHROME_K8M890:
            /* DFP Low Pad on */
            sr2a = read_reg(VIASR, SR2A);
            write_reg_mask(SR2A, VIASR, 0x03, BIT0+BIT1);
            break;

        case UNICHROME_P4M900:
        case UNICHROME_P4M890:
            /* DFP Low Pad on */
            sr2a = read_reg(VIASR, SR2A);
            write_reg_mask(SR2A, VIASR, 0x03, BIT0+BIT1);

            /* VT3364 also supports DVI on AMR, so we need to sense DVP0, too. */
            /* DVP0 Pad on */
            sr1e = read_reg(VIASR, SR1E);
            write_reg_mask(SR1E, VIASR, 0xC0, BIT6+BIT7);
            break;
               
        default:
            /* DVP0/DVP1 Pad on */
            sr1e = read_reg(VIASR, SR1E);
            write_reg_mask(SR1E, VIASR, 0xF0, BIT4+BIT5+BIT6+BIT7);
            /* SR3E[1]Multi-function selection:0 = Emulate I2C and DDC bus by GPIO2/3/4. */
            sr3e = read_reg(VIASR, SR3E);
            write_reg_mask(SR3E, VIASR, 0x0, BIT5);
            break;
    }

    /* Check for VT1632: */
    chip_info.tmds_chip_info.tmds_chip_name = VT1632_TMDS;
    chip_info.tmds_chip_info.tmds_chip_slave_addr = VT1632_TMDS_I2C_ADDR;
    chip_info.tmds_chip_info.i2c_port = TVI2CPORTINDEX;
    if (check_tmds_chip(VT1632_DEVICE_ID_REG, VT1632_DEVICE_ID)!= FAIL)
    {
        /* Currently only support 12bits, dual edge,add 24bits mode later */
        tmds_register_write(0x08, 0x3b);
        
        DEBUG_MSG(KERN_INFO "\n VT1632 TMDS ! \n");
        DEBUG_MSG(KERN_INFO "\n %2d", chip_info.tmds_chip_info.tmds_chip_name);
        DEBUG_MSG(KERN_INFO "\n %2d", chip_info.tmds_chip_info.i2c_port);
        return(OK);
    }else
    {
        chip_info.tmds_chip_info.i2c_port = TVGPIOPORTINDEX;
        if (check_tmds_chip(VT1632_DEVICE_ID_REG, VT1632_DEVICE_ID)!= FAIL)
        {
            /* Currently only support 12bits, dual edge,add 24bits mode later */
            tmds_register_write(0x08, 0x3b);
            
            DEBUG_MSG(KERN_INFO "\n VT1632 TMDS ! \n");
            DEBUG_MSG(KERN_INFO "\n %2d", chip_info.tmds_chip_info.tmds_chip_name);
            DEBUG_MSG(KERN_INFO "\n %2d", chip_info.tmds_chip_info.i2c_port);
            return(OK);
        }
    }

    /* Check for SII164: */
    chip_info.tmds_chip_info.tmds_chip_name = SII164_TMDS;
    chip_info.tmds_chip_info.tmds_chip_slave_addr = SII164_TMDS_I2C_ADDR;
    if (check_tmds_chip(SII164_DEVICE_ID_REG, SII164_DEVICE_ID)!= FAIL)
    {
        return(OK);
    }


    /* Restore status */
    switch (chip_info.gfx_chip_name)
    {
        case UNICHROME_K8M890:
            write_reg(SR2A, VIASR, sr2a);
            break;

        case UNICHROME_P4M900:
        case UNICHROME_P4M890:
            write_reg(SR2A, VIASR, sr2a);
            write_reg(SR1E, VIASR, sr1e);
            break;

        default: 
            write_reg(SR1E, VIASR, sr1e);
            write_reg(SR3E, VIASR, sr3e);
            break;
    }

    chip_info.tmds_chip_info.tmds_chip_name = NON_TMDS_TRANSMITTER;
    chip_info.tmds_chip_info.tmds_chip_slave_addr = VT1632_TMDS_I2C_ADDR;
    return(FAIL);
}

void tmds_register_write(int index, u8 data)
{
    u8  tmp;

    tmp = chip_info.tv_chip_info.tv_chip_on_slot;
    switch(chip_info.tmds_chip_info.i2c_port)
    {
        case TVI2CPORTINDEX:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AMR;
            break;
            
        case TVGPIOPORTINDEX:
        default:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AGP;
            break;
    }


    i2cWriteByte(chip_info.tmds_chip_info.tmds_chip_slave_addr, index, data);

    chip_info.tv_chip_info.tv_chip_on_slot = tmp;

}

int tmds_register_read(int index)
{
    u8 data;
    int status;
    u8  tmp;

    tmp = chip_info.tv_chip_info.tv_chip_on_slot;
    switch(chip_info.tmds_chip_info.i2c_port)
    {
        case TVI2CPORTINDEX:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AMR;
            break;
            
        case TVGPIOPORTINDEX:
        default:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AGP;
            break;
    }


    status = i2cReadByte((u8)chip_info.tmds_chip_info.tmds_chip_slave_addr, (u8)index, &data);
    chip_info.tv_chip_info.tv_chip_on_slot = tmp;
    return(data);
}

/* DVI Set Mode */
void dvi_set_mode(struct crt_mode_table *dvi_crt_timing, int video_index, int mode_bpp, int set_iga)
{
    fill_crtc_timing(dvi_crt_timing, video_index, mode_bpp/8, set_iga);
    set_output_path(DEVICE_DVI, set_iga, chip_info.tmds_chip_info.output_interface);
}

/* Sense DVI Connector */
int dvi_sense(void)
{
    u8  RegSR1E=0, RegSR3E=0, RegCR6B=0, RegCR91=0, RegCR93=0, RegCR9B=0, data;
    int ret = FALSE;

    DEBUG_MSG(KERN_INFO "dvi_sense!!\n");

    if (chip_info.gfx_chip_name == UNICHROME_CLE266)
    {
        /* DI1 Pad on */
        RegSR1E = read_reg(VIASR, SR1E);
        write_reg(SR1E, VIASR, RegSR1E|0x30);

        /* CR6B[0]VCK Input Selection: 1 = External clock. */
        RegCR6B = read_reg(VIACR, CR6B);
        write_reg(CR6B, VIACR, RegCR6B|0x08);

        /* CR91[4] VDD On [3] Data On [2] VEE On [1] Back Light Off
           [0] Software Control Power Sequence */
        RegCR91 = read_reg(VIACR, CR91);
        write_reg(CR91, VIACR, 0x1D);

        /* CR93[7] DI1 Data Source Selection: 1 = DSP2.
           CR93[5] DI1 Clock Source: 1 = internal.
           CR93[4] DI1 Clock Polarity.
           CR93[3:1] DI1 Clock Adjust. CR93[0] DI1 enable */
        RegCR93 = read_reg(VIACR, CR93);
        write_reg(CR93, VIACR, 0x01);
    }
    else
    {
        /* DVP0/DVP1 Pad on */
        RegSR1E = read_reg(VIASR, SR1E);
        write_reg(SR1E, VIASR, RegSR1E|0xF0);

        /* SR3E[1]Multi-function selection:
                    0 = Emulate I2C and DDC bus by GPIO2/3/4. */
        RegSR3E = read_reg(VIASR, SR3E);
        write_reg(SR3E, VIASR, RegSR3E&(~0x20));

        /* CR91[4] VDD On [3] Data On [2] VEE On [1] Back Light Off
               [0] Software Control Power Sequence */
        RegCR91 = read_reg(VIACR, CR91);
        write_reg(CR91, VIACR, 0x1D);

        /* CR9B[4] DVP1 Data Source Selection: 1 = From secondary display.
             CR9B[2:0] DVP1 Clock Adjust */
        RegCR9B = read_reg(VIACR, CR9B);
        write_reg(CR9B, VIACR, 0x01);
    }

    data = (u8)tmds_register_read(0x09);
    if(data & 0x04)
        ret = TRUE;

    if (ret == FALSE)
    {
        if(dvi_query_EDID())
        {
            ret = TRUE;
        }
    }

    /* Restore status */
    if (chip_info.gfx_chip_name == UNICHROME_CLE266)
    {
        write_reg(SR1E, VIASR, RegSR1E);
        write_reg(CR6B, VIACR, RegCR6B);
        write_reg(CR91, VIACR, RegCR91);
        write_reg(CR93, VIACR, RegCR93);
    }
    else
    {
        write_reg(SR1E, VIASR, RegSR1E);
        write_reg(SR3E, VIASR, RegSR3E);
        write_reg(CR91, VIACR, RegCR91);
        write_reg(CR9B, VIACR, RegCR9B);
    }

    return ret;
}

/* Query Flat Panel's EDID Table Version Through DVI Connector */
int dvi_query_EDID(void)
{
    u8  data0, data1;

    DEBUG_MSG(KERN_INFO "dvi_query_EDID!!\n");

    data0 = (u8)tmds_register_read(0x00);
    data1 = (u8)tmds_register_read(0x01);
    if ((data0 == 0) && (data1 == 0xFF))
        return EDID_VERSION_1; /* Found EDID1 Table */

    data0 = (u8)tmds_register_read(0x00);
    if (data0 == 0x20)
        return EDID_VERSION_2; /* Found EDID2 Table */
    else
        return 0;
}

/* If Disable DVI, turn off pad */
void dvi_disable(void)
{
    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DVP0)
    {
        write_reg(SR1E, VIASR, read_reg(VIASR, SR1E) & (~0xC0));
    }

    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DVP1)
    {
        write_reg(SR1E, VIASR, read_reg(VIASR, SR1E) & (~0x30));
    }

    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DFP_HIGH)
    {
        write_reg(SR2A, VIASR, read_reg(VIASR, SR2A) & (~0x0C));
    }

    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DFP_LOW)
    {
        write_reg(SR2A, VIASR, read_reg(VIASR, SR2A) & (~0x03));
    }

    /* Add for VT3324 integrated TMDS. */
    if (chip_info.tmds_chip_info.output_interface == INTERFACE_TMDS)
    {
        write_reg(CRD2, VIACR, read_reg(VIACR, CRD2) | 0x08);   /* Turn off TMDS power. */
    }
}

/* If Enable DVI, turn off pad */
void dvi_enable(void)
{
    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DVP0)
    {
        write_reg(SR1E, VIASR, read_reg(VIASR, SR1E) | 0xC0);
        if(chip_info.gfx_chip_name == UNICHROME_CLE266)
            tmds_register_write(0x88,0x3b);
        else
            /* Clear CR91[5] to direct on display period in the secondary diplay path */
            write_reg(CR91,VIACR,read_reg(VIACR,CR91) & 0xDF);
    }

    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DVP1)
    {
        write_reg(SR1E, VIASR, read_reg(VIASR, SR1E) | 0x30);

        /* Fix dvi cann't be enabled with MB VT5718C4 */
        if(chip_info.gfx_chip_name == UNICHROME_CLE266)
        {   
            tmds_register_write(0x88,0x3b);
        }
        else
        {
            /* Clear CR91[5] to direct on display period in the secondary diplay path */
            write_reg(CR91,VIACR,read_reg(VIACR,CR91) & 0xDF);
        }
        
        /* Fix DVI cannot enable on EPIA-M board */
        if(EPIA_DVI==1)
        {
                write_reg_mask(CR91,VIACR,0x1f,0x1f);
                write_reg_mask(CR88,VIACR,0x00,BIT6+BIT0);                
           if(BusWidth==24)
           {
    	        if(LCDDualEdge==1)
    	            i2cWriteByte(chip_info.tmds_chip_info.tmds_chip_slave_addr, 0x08, 0x3F);
    	        else
    	            i2cWriteByte(chip_info.tmds_chip_info.tmds_chip_slave_addr, 0x08, 0x37);
            }
        }
    }
    
    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DFP_HIGH)
    {
        write_reg(SR2A, VIASR, read_reg(VIASR, SR2A) | 0x0C);
        
        /* Fix can't light on DVI problem
           clear CR91[5] to direct on display period in the secondary diplay path */
        write_reg(CR91,VIACR,read_reg(VIACR,CR91) & 0xDF);
    }

    if (chip_info.tmds_chip_info.output_interface == INTERFACE_DFP_LOW)
    {
        write_reg(SR2A, VIASR, read_reg(VIASR, SR2A) | 0x03);
        
        /* Fix can't light on DVI problem
           clear CR91[5] to direct on display period in the secondary diplay path */
        write_reg(CR91,VIACR,read_reg(VIACR,CR91) & 0xDF);
    }

    /* Add for VT3324 integrated TMDS. */
    if (chip_info.tmds_chip_info.output_interface == INTERFACE_TMDS)
    {
        /* Turn on Display period in the panel path. */
        write_reg_mask(CR91, VIACR, 0, BIT7);

        /* Turn on TMDS power. */
        write_reg_mask(CRD2, VIACR, 0, BIT3);
    }
}

