 /*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 

#include <linux/ioport.h>
#include <linux/pci.h>
#include <asm/io.h>

#include "iface.h"
#include "debug.h"
#include "share.h"

/* Get frame buffer size from VGA BIOS */
extern u8 read_reg(int io_port, u8 index);
extern  int memsize;

unsigned int get_memsize(void)
{
    unsigned int m;

        /* If memory size provided by user */
    if (memsize)
        m = memsize * Mb;
    else
    {
        m = (unsigned int) read_reg(VIASR, SR39); 
        m = m*(4*Mb);
        
        if((m < (16*Mb)) || (m > (64*Mb)))
            m = 16*Mb;
    }
    DEBUG_MSG(KERN_INFO "framebuffer size = %d Mb\n", m/Mb);
    return m;
}

/* Get Video Buffer Starting Physical Address(back door) */

unsigned long get_videobuf_addr(void)
{
    struct pci_dev *pdev=NULL;
    unsigned char  sys_mem;
    unsigned char  video_mem;
    unsigned long  sys_mem_size;
    unsigned long  video_mem_size;
    unsigned long  vmem_starting_adr = 0x0C000000; /* system memory = 256 MB, video memory 64 MB */

    pdev = (struct pci_dev *)pci_find_device(VIA_K800_BRIDGE_VID, VIA_K800_BRIDGE_DID, NULL);
    if (pdev != NULL)
    {
        pci_read_config_byte(pdev, VIA_K800_SYSTEM_MEMORY_REG, &sys_mem);
        pci_read_config_byte(pdev, VIA_K800_VIDEO_MEMORY_REG, &video_mem);
        video_mem = (video_mem & 0x70) >> 4;
        sys_mem_size = ((unsigned long)sys_mem) << 24;
        if (video_mem != 0)
            video_mem_size = (1 << (video_mem)) *1024*1024;
        else
            video_mem_size = 0;

        vmem_starting_adr = sys_mem_size - video_mem_size;

    }

    DEBUG_MSG(KERN_INFO "Video Memory Starting Address = %lx \n", vmem_starting_adr);
    return vmem_starting_adr;
}




