/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <linux/fb.h>
#include <asm/uaccess.h>

#include "viafbdev.h"
#include "chip.h"
#include "debug.h"
#include "ioctl.h"
#include "share.h"

extern int dvi_sense(void);
extern int tv_sense(void);
extern int tv_auto_sense(void);
extern inline u8 read_reg(int io_port, u8 index);

extern struct chip_information          chip_info;
extern struct tv_setting_information    tv_setting_info;
extern struct viafb_info                viafbinfo;
extern int CRT_ON, TV_ON, DVI_ON, LCD_ON;
extern int DeviceStatus;
extern int HotPlug_TV;
extern int refresh;

int HotPlug_Xres = 640, HotPlug_Yres = 480, HotPlug_bpp = 32, HotPlug_refresh = 60, HotPlug_TVType = TVTYPE_NTSC;

int ioctl_get_viafb_info(u_long arg)
{
    struct viafb_ioctl_info viainfo;

    viainfo.viafb_id    = VIAID;
    viainfo.vendor_id   = PCI_VIA_VENDOR_ID;

    switch(chip_info.gfx_chip_name) {
        case UNICHROME_CLE266:
            viainfo.device_id = UNICHROME_CLE266_DID;
            break;

        case UNICHROME_K400:
            viainfo.device_id = UNICHROME_K400_DID;
            break;

        case UNICHROME_K800:
            viainfo.device_id = UNICHROME_K800_DID;
            break;

        case UNICHROME_P880:
            viainfo.device_id = UNICHROME_P880_DID;
            break;

        case UNICHROME_CN900:
            viainfo.device_id = UNICHROME_CN900_DID;
            break;
            
        /* Add for VT3324. */
        case UNICHROME_CX700:
            viainfo.device_id = UNICHROME_CX700_DID;
            break;

        /* Add for VT3327 VT3336 VT3364 */
        case UNICHROME_K8M890:
            viainfo.device_id = UNICHROME_K8M890_DID;
            break;

        case UNICHROME_P4M890:
            viainfo.device_id = UNICHROME_P4M890_DID;
            break;

        case UNICHROME_P4M900:
            viainfo.device_id = UNICHROME_P4M900_DID;
            break;
    }
    
    viainfo.version = VERSION_MAJOR;
    viainfo.revision = VERSION_MINOR;

    if(copy_to_user((void __user *)arg, &viainfo, sizeof(viainfo))) {
        return -EFAULT;
    }

    return 0;
}

/* Hot-Plug Priority: DVI > TV > CRT */
int ioctl_hotplug(int hres, int vres, int bpp)
{
    int TVsense, DVIsense, status = 0;
    u8  TVType = 0;

    DEBUG_MSG(KERN_INFO "ioctl_hotplug!!\n");

    if(chip_info.tmds_chip_info.tmds_chip_name != NON_TMDS_TRANSMITTER) {
        DVIsense = dvi_sense();

        if(DVIsense) {
            DEBUG_MSG(KERN_INFO "DVI Attached...\n");
            if(DeviceStatus != DVI_Device) {
                DVI_ON = 1;
                CRT_ON = TV_ON = LCD_ON = 0;

                if(DeviceStatus == TV_Device) {
                    tv_setting_info.system = HotPlug_TVType;
                }

                DeviceStatus = DVI_Device;
                return DeviceStatus;
            }
            status = 1;
        }
        else
            DEBUG_MSG(KERN_INFO "DVI De-attached...\n");
    }
    
    if((chip_info.tv_chip_info.tv_chip_name != NON_TV_ENCODER)&&(status == 0)) {
        /* If TV not ON, using TV manual sense function.  
           If TV ON, using TV auto sense function. 
           This is avoid the display flick to use TV manual sense function when TV ON. */
        if(DeviceStatus != TV_Device) {
            TVsense = tv_sense();
        }
        else {
            TVsense = tv_auto_sense();
        }

        if(TVsense != TV_OUTPUT_NONE) {
            DEBUG_MSG(KERN_INFO "TV Attached: TVsense = %d\n", TVsense);
    
            if(DeviceStatus != TV_Device) {
                TV_ON = 1;
                CRT_ON = DVI_ON = LCD_ON = 0;

                HotPlug_TVType = tv_setting_info.system;

                if((chip_info.gfx_chip_name == UNICHROME_CLE266) && (chip_info.gfx_chip_revision == CLE266_REVISION_AX)) {
                    TVType = read_reg(VIACR, 0x3C) & 0x0F;
                }
                else {
                    TVType = read_reg(VIACR, 0x4E) & 0x0F;
                }

                if(!TVType) {
                    tv_setting_info.system = TVTYPE_NTSC;
                }
                else {
                    tv_setting_info.system = TVTYPE_PAL;
                }

                tv_setting_info.out_signal = TVsense;

                HotPlug_TV = 1;
                DeviceStatus = TV_Device;
                return DeviceStatus;
            }
            status = 1;
        }
        else {
            DEBUG_MSG(KERN_INFO "TV De-attached...\n");
            HotPlug_TV = 0;
        }
    }
    
    if((DeviceStatus != CRT_Device)&&(status == 0)) {
        CRT_ON = 1;
        TV_ON = DVI_ON = LCD_ON = 0;

        if(DeviceStatus == TV_Device) {
            tv_setting_info.system = HotPlug_TVType;
        }

        DeviceStatus = CRT_Device;
        return DeviceStatus;
    }

    return 0;
}

int ioctl_get_tv_type(u_long arg)
{
    u8          TVType = 0;
    u32 __user  *argp = (u32 __user *)arg;

    if((chip_info.gfx_chip_name == UNICHROME_CLE266) && (chip_info.gfx_chip_revision == CLE266_REVISION_AX)) {
        TVType = read_reg(VIACR, 0x3C) & 0x0F;
    }
    else {
        TVType = read_reg(VIACR, 0x4E) & 0x0F;
    }

    return put_user((u32)TVType, argp);
}

