/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef _TBLDPASETTING_H_
#define _TBLDPASETTING_H_ 1
#include "lcd.h"

#define DPA_CLK_30M       30000000
#define DPA_CLK_50M       50000000
#define DPA_CLK_70M       70000000
#define DPA_CLK_100M      100000000
#define DPA_CLK_150M      150000000


typedef enum _DPA_CLK_RANGE {
    DPA_CLK_RANGE_30M,
    DPA_CLK_RANGE_30_50M,
    DPA_CLK_RANGE_50_70M,
    DPA_CLK_RANGE_70_100M,
    DPA_CLK_RANGE_100_150M,
    DPA_CLK_RANGE_150M,
} DPA_CLK_RANGE;


/* For VT3324: */
struct VT1636_DPA_SETTING VT1636_DPA_SETTING_TBL_VT3324[] = {
     /* Panel ID,       CLK_SEL_ST1[09],    CLK_SEL_ST2[08] */
    {LCD_PANEL_ID0_640X480,      0x00,               0x00},          /* For 640x480   */
    {LCD_PANEL_ID1_800X600,      0x00,               0x00},          /* For 800x600   */
    {LCD_PANEL_ID2_1024X768,     0x00,               0x00},          /* For 1024x768  */
    {LCD_PANEL_ID3_1280X768,     0x00,               0x00},          /* For 1280x768  */
    {LCD_PANEL_ID4_1280X1024,    0x00,               0x00},          /* For 1280x1024 */
    {LCD_PANEL_ID5_1400X1050,    0x00,               0x00},          /* For 1400x1050 */
    {LCD_PANEL_ID6_1600X1200,    0x0B,               0x03}           /* For 1600x1200 */
};

struct GFX_DPA_SETTING GFX_DPA_SETTING_TBL_VT3324[] = {
    /*  ClkRange,            DVP0, DVP0DataDriving,  DVP0ClockDriving, DVP1, DVP1Driving, DFPHigh,   DFPLow */
    /*                       CR96, SR2A[5], SR1B[1], SR2A[4], SR1E[2], CR9B, SR65,        CR97,      CR99   */
    {DPA_CLK_RANGE_30M,      0x00, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x00},  /* LCK/VCK < 30000000 will use this value */
    {DPA_CLK_RANGE_30_50M,   0x00, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x00},  /* 30000000 < LCK/VCK < 50000000 will use this value */
    {DPA_CLK_RANGE_50_70M,   0x00, 0x00,    0x00,    0x00,    0x00,    0x02, 0x00,        0x00,      0x00},  /* 50000000 < LCK/VCK < 70000000 will use this value */
    {DPA_CLK_RANGE_70_100M,  0x00, 0x00,    0x00,    0x00,    0x00,    0x02, 0x00,        0x00,      0x00},  /* 70000000 < LCK/VCK < 100000000 will use this value */
    {DPA_CLK_RANGE_100_150M, 0x00, 0x00,    0x00,    0x00,    0x00,    0x02, 0x00,        0x00,      0x00},  /* 100000000 < LCK/VCK < 15000000 will use this value */
    {DPA_CLK_RANGE_150M,     0x00, 0x00,    0x00,    0x00,    0x00,    0x08, 0x0E,        0x00,      0x00},  /* 15000000 < LCK/VCK will use this value */
};


/* For VT3327: */
struct VT1636_DPA_SETTING VT1636_DPA_SETTING_TBL_VT3327[] = {
     /* Panel ID,       CLK_SEL_ST1[09],    CLK_SEL_ST2[08] */
    {LCD_PANEL_ID0_640X480,      0x00,               0x00},          /* For 640x480   */
    {LCD_PANEL_ID1_800X600,      0x00,               0x00},          /* For 800x600   */
    {LCD_PANEL_ID2_1024X768,     0x00,               0x00},          /* For 1024x768  */
    {LCD_PANEL_ID3_1280X768,     0x00,               0x00},          /* For 1280x768  */
    {LCD_PANEL_ID4_1280X1024,    0x00,               0x00},          /* For 1280x1024 */
    {LCD_PANEL_ID5_1400X1050,    0x00,               0x00},          /* For 1400x1050 */
    {LCD_PANEL_ID6_1600X1200,    0x00,               0x00}           /* For 1600x1200 */
};

struct GFX_DPA_SETTING GFX_DPA_SETTING_TBL_VT3327[] = {
    /*  ClkRange,            DVP0, DVP0DataDriving,  DVP0ClockDriving, DVP1, DVP1Driving, DFPHigh,   DFPLow */
    /*                       CR96, SR2A[5], SR1B[1], SR2A[4], SR1E[2], CR9B, SR65,        CR97,      CR99   */
    {DPA_CLK_RANGE_30M,      0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x08,      0x01},  /* LCK/VCK < 30000000 will use this value */
    {DPA_CLK_RANGE_30_50M,   0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x08,      0x01},  /* 30000000 < LCK/VCK < 50000000 will use this value */
    {DPA_CLK_RANGE_50_70M,   0x06, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x08,      0x01},  /* 50000000 < LCK/VCK < 70000000 will use this value */
    {DPA_CLK_RANGE_70_100M,  0x03, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x08,      0x03},  /* 70000000 < LCK/VCK < 100000000 will use this value */
    {DPA_CLK_RANGE_100_150M, 0x03, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x01,      0x02},  /* 100000000 < LCK/VCK < 15000000 will use this value */
    {DPA_CLK_RANGE_150M,     0x00, 0x20,    0x00,    0x10,    0x00,    0x03, 0x00,        0x0D,      0x03},  /* 15000000 < LCK/VCK will use this value */
};


/* For VT3364: */
struct GFX_DPA_SETTING GFX_DPA_SETTING_TBL_VT3364[] = {
    /*  ClkRange,            DVP0, DVP0DataDriving,  DVP0ClockDriving, DVP1, DVP1Driving, DFPHigh,   DFPLow */
    /*                       CR96, SR2A[5], SR1B[1], SR2A[4], SR1E[2], CR9B, SR65,        CR97,      CR99   */
    {DPA_CLK_RANGE_30M,      0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x08},  /* LCK/VCK < 30000000 will use this value */
    {DPA_CLK_RANGE_30_50M,   0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x08},  /* 30000000 < LCK/VCK < 50000000 will use this value */
    {DPA_CLK_RANGE_50_70M,   0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x08},  /* 50000000 < LCK/VCK < 70000000 will use this value */
    {DPA_CLK_RANGE_70_100M,  0x07, 0x00,    0x00,    0x00,    0x00,    0x03, 0x00,        0x00,      0x08},  /* 70000000 < LCK/VCK < 100000000 will use this value */
    {DPA_CLK_RANGE_100_150M, 0x03, 0x00,    0x02,    0x00,    0x00,    0x03, 0x00,        0x00,      0x08},  /* 100000000 < LCK/VCK < 15000000 will use this value */
    {DPA_CLK_RANGE_150M,     0x01, 0x00,    0x02,    0x10,    0x00,    0x03, 0x00,        0x00,      0x08},  /* 15000000 < LCK/VCK will use this value */
};

#endif

