/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef __VIAFBDEV_H__
#define __VIAFBDEV_H__

#include <linux/version.h>
#include <linux/proc_fs.h>
#define VERSION_MAJOR       2

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#define VERSION_KERNEL      4   /* For kernel 2.4 */
#else
#define VERSION_KERNEL      6   /* For kernel 2.6 */
#endif

#define VERSION_OS          0   /* 0: for 32 bits OS, 1: for 64 bits OS */
#define VERSION_MINOR       3

struct viafb_par {
    struct fb_var_screeninfo var;
    int bpp;
    int hres;
    int vres;
    int linelength;
    int vclk;       /* in MHz */

    int vtotal;
    int vdispend;
    int vsyncstart;
    int vsyncend;
    int vblankstart;
    int vblankend;

    int htotal;
    int hdispend;
    int hsyncstart;
    int hsyncend;
    int hblankstart;
    int hblankend;

    int accel;                  /* Acceleration Flags */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
    void __iomem *fbmem_virt;   /* framebuffer virtual memory address */
    void __iomem *io_virt;      /* iospace virtual memory address */
    void __iomem *integrated_tv_map_base;   /* Integrated TV start address */
#else
    unsigned int fbmem_virt;    /* framebuffer virtual memory address */
    unsigned int io_virt;       /* iospace virtual memory address */
    unsigned int integrated_tv_map_base;    /* Integrated TV start address */
#endif
    unsigned int fbmem;         /* framebuffer physical memory address */
    unsigned int memsize;       /* size of fbmem */
    unsigned int io;            /* io space address */
    unsigned long mmio_base;    /* mmio base address */
    unsigned long mmio_len;     /* mmio base length */
    u32 fbmem_free;             /* Free FB memory */
    u32 fbmem_used;             /* Use FB memory size */
    u32 cursor_start;           /* Cursor Start Address */
    u32 VQ_start;               /* Virtual Queue Start Address */
    u32 VQ_end;                 /* Virtual Queue End Address */
    u32 iga_path;               /* For dual-fb, see hw.c:set_start_addr */
#endif
    u8 duoview;  /* Is working in duoview mode? */
    struct proc_dir_entry *proc_entry; /* viafb proc entry */
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
struct viafb_info {
    struct fb_info_gen gen;
    unsigned int fbmem_virt;    /* framebuffer virtual memory address */
    unsigned int fbmem;         /* framebuffer physical memory address */
    unsigned int memsize;       /* size of fbmem */
    unsigned int io;            /* io space address */
    unsigned int io_virt;       /* iospace virtual memory address */
    unsigned long mmio_base;    /* mmio base address */
    unsigned long mmio_len;     /* mmio base length */
    struct viafb_par currentmode;
    unsigned char eng_oper;     /* engine operation... */
    unsigned int integrated_tv_map_base;    /* Integrated TV start address */
};
#endif



#endif /* __VIAFBDEV_H__ */
