/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef __VIAMODE_H__
#define __VIAMODE_H__

#include "share.h"


struct VPITTable {
  unsigned char  Misc;
  unsigned char  SR[StdSR];
  unsigned char  GR[StdGR];
  unsigned char  AR[StdAR];
};

struct VideoModeTable {
  int                               ModeIndex;
  struct crt_mode_table             *crtc;
  int                               mode_array;
};

struct patch_table {
  int           mode_index;
  int           table_length;
  struct io_reg *io_reg_table;
};

struct res_map_refresh {
  int       hres;
  int       vres;
  int       pixclock;
  int       vmode_refresh;
};


/********************/
/* Mode Table       */
/********************/

/* 640x480 */
static struct crt_mode_table CRTM640x480[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60, CLK_25_175M ,M640X480_R60_HSP, M640X480_R60_VSP,\
  {800, 640, 648, 144, 656, 96,  525, 480, 480, 45,  490, 2}},
  {REFRESH_75, CLK_31_500M ,M640X480_R75_HSP, M640X480_R75_VSP,\
  {840, 640, 640, 200, 656, 64,  500, 480, 480, 20,  481, 3}},
  {REFRESH_85, CLK_36_000M ,M640X480_R85_HSP, M640X480_R85_VSP,\
  {832, 640, 640, 192, 696, 56,  509, 480, 480, 29,  481, 3}},
  {REFRESH_100,CLK_43_163M ,M640X480_R100_HSP, M640X480_R100_VSP,\
  {848, 640, 640, 208, 680, 64,  509, 480, 480, 29,  481, 3}}, /* GTF */
  {REFRESH_120,CLK_52_406M ,M640X480_R120_HSP, M640X480_R120_VSP,\
  {848, 640, 640, 208, 680, 64,  515, 480, 480, 35,  481, 3}}  /* GTF */
};

/* 720x480 (GTF) */
static struct crt_mode_table CRTM720x480[] = {
  /* r_rate,vclk,hsp,vsp */     
  /* HT, HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_26_880M ,M720X480_R60_HSP, M720X480_R60_VSP,\
  {896, 720, 720, 176, 736, 72,  497, 480, 480, 17,  481, 3}}

};

/* 720x576 (GTF) */
static struct crt_mode_table CRTM720x576[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_32_668M ,M720X576_R60_HSP, M720X576_R60_VSP,\
  {912, 720, 720, 192, 744, 72,  597, 576, 576, 21,  577, 3}}
};

/* 800x480 (GTF) */
static struct crt_mode_table CRTM800x480[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_29_581M ,M800X480_R60_HSP,M800X480_R60_VSP,\
  {992, 800, 800, 192, 816, 80,  497, 480, 480, 17,  481, 3}}
};

/* 800x600 */
static struct crt_mode_table CRTM800x600[] = {
  /* r_rate,vclk,hsp,vsp */     
  /* HT,   HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60, CLK_40_000M ,M800X600_R60_HSP, M800X600_R60_VSP,\
  {1056, 800, 800, 256, 840, 128, 628, 600, 600, 28,  601, 4}},
  {REFRESH_75, CLK_49_500M ,M800X600_R75_HSP, M800X600_R75_VSP,\
  {1056, 800, 800, 256, 816, 80,  625, 600, 600, 25,  601, 3}},
  {REFRESH_85, CLK_56_250M ,M800X600_R85_HSP, M800X600_R85_VSP,\
  {1048, 800, 800, 248, 832, 64,  631, 600, 600, 31,  601, 3}},
  {REFRESH_100,CLK_68_179M ,M800X600_R100_HSP, M800X600_R100_VSP,\
  {1072, 800, 800, 272, 848, 88,  636, 600, 600, 36,  601, 3}},
  {REFRESH_120,CLK_83_950M ,M800X600_R120_HSP, M800X600_R120_VSP,\
  {1088, 800, 800, 288, 856, 88,  643, 600, 600, 43,  601, 3}}
};

/* 848x480 (GTF) */
static struct crt_mode_table CRTM848x480[] = {
  /* r_rate,vclk,hsp,vsp */     
  /* HT,   HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_31_490M ,M848X480_R60_HSP, M848X480_R60_VSP,\
  {1056, 848, 848, 208, 864, 88,  497, 480, 480, 17,  481, 3}}
};

/* 856x480 (GTF) convert to 852x480 */
static struct crt_mode_table CRTM852x480[] = {
  /* r_rate,vclk,hsp,vsp */    
  /* HT,   HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_31_728M ,M852X480_R60_HSP, M852X480_R60_VSP,\
  {1064, 856, 856, 208, 872, 88,  497, 480, 480, 17,  481, 3}}

};

/* 1024x512 (GTF) */
static struct crt_mode_table CRTM1024x512[] = {
  /* r_rate,vclk,hsp,vsp */    
  /* HT,   HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_41_291M ,M1024X512_R60_HSP, M1024X512_R60_VSP,\
  {1296, 1024,1024,272, 1056,104, 531, 512, 512, 19,  513, 3}}

};

/* 1024x576 (GTF) */
/*static struct crt_mode_table CRTM1024x576[] = {
  //r_rate,vclk,     HT,   HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE
  { 60,CLK_46_996M ,{1312, 1024,1024,288, 1064,104, 597, 576, 576, 21,  577, 3}}

};*/

/* 1024x768 */
static struct crt_mode_table CRTM1024x768[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_65_000M ,M1024X768_R60_HSP, M1024X768_R60_VSP,\
  {1344, 1024, 1024, 320, 1048, 136, 806, 768, 768, 38,  771, 6}},
  {REFRESH_75,CLK_78_750M ,M1024X768_R75_HSP, M1024X768_R75_VSP,\
  {1312, 1024, 1024, 288, 1040, 96,  800, 768, 768, 32,  769, 3}},
  {REFRESH_85,CLK_94_500M ,M1024X768_R85_HSP, M1024X768_R85_VSP,\
  {1376, 1024, 1024, 352, 1072, 96,  808, 768, 768, 40,  769, 3}},
  {REFRESH_100,CLK_133_308M,M1024X768_R100_HSP, M1024X768_R100_VSP,\
  {1392, 1024, 1024, 368, 1096, 112, 814, 768, 768, 46,  769, 3}} //GTF
};

/* 1152x864 */
static struct crt_mode_table CRTM1152x864[] = {
  /* r_rate,vclk,hsp,vsp */      
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_75,CLK_108_000M ,M1152X864_R75_HSP, M1152X864_R75_VSP,\
  {1600, 1152,1152, 448, 1216, 128, 900, 864, 864, 36,  865, 3}}

};

/* 1280x720 (GTF) */
static struct crt_mode_table CRTM1280x720[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */      
  {REFRESH_60,CLK_74_481M ,M1280X720_R60_HSP, M1280X720_R60_VSP,\
  {1664,1280, 1280, 384, 1336, 136, 746, 720, 720, 26,  721, 3}}
};

/* 1280x768 (GTF) */
static struct crt_mode_table CRTM1280x768[] = {
  /* r_rate,vclk,hsp,vsp */     
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_80_136M ,M1280X768_R60_HSP, M1280X768_R60_VSP,\
  {1680,1280, 1280, 400, 1344, 136, 795, 768, 768, 27,  769, 3}}
};

/* 1280x800 (GTF) */
static struct crt_mode_table CRTM1280x800[] = {
  /* r_rate,          vclk,              hsp,             vsp   */
  /* HT,  HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60, CLK_83_375M, M1280X800_R60_HSP, M1280X800_R60_VSP,\
  {1680,1280, 1280, 400, 1344, 136, 828, 800, 800, 28,  801, 3}}
};

/* 1280x960 */
static struct crt_mode_table CRTM1280x960[] = {
  /* r_rate,vclk,hsp,vsp */ 
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_108_000M ,M1280X960_R60_HSP, M1280X960_R60_VSP,\
  {1800,1280, 1280, 520, 1376, 112, 1000,960, 960, 40,  961, 3}}  
};

/* 1280x1024 */
static struct crt_mode_table CRTM1280x1024[] = {
  /* r_rate,vclk,,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_108_000M ,M1280X1024_R60_HSP, M1280X1024_R60_VSP,\
  {1688,1280, 1280, 408, 1328, 112, 1066,1024,1024,42,  1025,3}},
  {REFRESH_75,CLK_135_000M ,M1280X1024_R75_HSP, M1280X1024_R75_VSP,\
  {1688,1280, 1280, 408, 1296, 144, 1066,1024,1024,42,  1025,3}},
  {REFRESH_85,CLK_157_500M ,M1280X1024_R85_HSP, M1280X1024_R85_VSP,\
  {1728,1280, 1280, 448, 1344, 160, 1072,1024,1024,48,  1025,3}}
};

/* 1368x768 (GTF) */
static struct crt_mode_table CRTM1368x768[] = {
  /* r_rate,  vclk, hsp, vsp */
  /* HT,  HA,  HBS, HBE, HSS, HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60, CLK_85_860M, M1368X768_R60_HSP, M1368X768_R60_VSP,\
  {1800,1368, 1368, 432, 1440, 144, 795, 768, 768, 27,  769, 3}}
};

/* 1368x768 (GTF) */
/*static struct crt_mode_table CRTM1368x768[] = {
  //r_rate,vclk,     HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE 
  { 60,CLK_85_860M ,{1800,1368, 1368, 432, 1440, 144, 795, 768, 768, 27,  769, 3}}
};*/

/* 1440x1050 (GTF) */
static struct crt_mode_table CRTM1440x1050[] = {
  /* r_rate,vclk,hsp,vsp */       
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60 ,CLK_125_104M ,M1440X1050_R60_HSP, M1440X1050_R60_VSP,\
  {1936,1440, 1440, 496, 1536, 152, 1077,1040,1040,37,  1041,3}}
};

/* 1600x1200 */
static struct crt_mode_table CRTM1600x1200[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */      
  {REFRESH_60 ,CLK_162_000M ,M1600X1200_R60_HSP, M1600X1200_R60_VSP,\
  {2160,1600, 1600, 560, 1664, 192, 1250,1200,1200,50,  1201,3}},
  {REFRESH_75 ,CLK_202_500M ,M1600X1200_R75_HSP, M1600X1200_R75_VSP,\
  {2160,1600, 1600, 560, 1664, 192, 1250,1200,1200,50,  1201,3}}

};

/* 1920x1080 (GTF) */
static struct crt_mode_table CRTM1920x1080[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */                            
  {REFRESH_60,CLK_172_798M ,M1920X1080_R60_HSP, M1920X1080_R60_VSP,\
  {2576,1920, 1920, 656, 2040, 208, 1118,1080,1080,38, 1081, 3}}
};

/* 1920x1440 */
static struct crt_mode_table CRTM1920x1440[] = {
  /* r_rate,vclk,hsp,vsp */
  /* HT,  HA,   HBS,  HBE, HSS,  HSE, VT,  VA,  VBS, VBE, VSS, VSE */
  {REFRESH_60,CLK_234_000M ,M1920X1440_R60_HSP, M1920X1440_R60_VSP,\
  {2600,1920, 1920, 680, 2048, 208, 1500,1440,1440,60,  1441,3}},
  {REFRESH_75,CLK_297_500M ,M1920X1440_R75_HSP, M1920X1440_R75_VSP,\
  {2640,1920, 1920, 720, 2064, 224, 1500,1440,1440,60,  1441,3}}
};

#if 0
/* 1400x1050 (VESA) */
static struct crt_mode_table CRTM1400x1050[] = {
  /* r_rate,          vclk,              hsp,             vsp   */
  /* HT,  HA,  HBS, HBE, HSS, HSE,    VT,  VA,  VBS, VBE,  VSS, VSE */
  {REFRESH_60,CLK_108_000M, M1400X1050_R60_HSP, M1400X1050_R60_VSP,
  {1688,1400, 1400, 288, 1448, 112, 1066,1050, 1050,  16, 1051, 3}}
};
#endif

/* 1400x1050 (GTF) */
static struct crt_mode_table CRTM1400x1050[] = {
  /* r_rate,          vclk,              hsp,             vsp   */
  /* HT,  HA,  HBS, HBE,  HSS, HSE,   VT,  VA,  VBS, VBE,  VSS, VSE */
  {REFRESH_60,CLK_122_614M, M1400X1050_R60_HSP, M1400X1050_R60_VSP,
  {1880,1400, 1400, 480, 1488, 152, 1087,1050, 1050,  37, 1051, 3}}
};

static struct VideoModeTable CLE266Modes[] = {
   /* Display : 640x480 */
   { VIA_RES_640X480,  CRTM640x480, ARRAY_SIZE(CRTM640x480)},

   /* Display : 720x480 (GTF) */
   { VIA_RES_720X480,  CRTM720x480, ARRAY_SIZE(CRTM720x480)},

   /* Display : 720x576 (GTF) */
   { VIA_RES_720X576,  CRTM720x576, ARRAY_SIZE(CRTM720x576)},

   /* Display : 800x600 */
   { VIA_RES_800X600,  CRTM800x600, ARRAY_SIZE(CRTM800x600)},

   /* Display : 800x480 (GTF) */
   { VIA_RES_800X480,  CRTM800x480, ARRAY_SIZE(CRTM800x480)},

   /* Display : 848x480 (GTF) */
   { VIA_RES_848X480,  CRTM848x480, ARRAY_SIZE(CRTM848x480)},

   /* Display : 852x480 (GTF) */
   { VIA_RES_856X480,  CRTM852x480, ARRAY_SIZE(CRTM852x480)},

   /* Display : 1024x512 (GTF) */
   { VIA_RES_1024X512, CRTM1024x512, ARRAY_SIZE(CRTM1024x512)},

    /* Display : 1024x576 (GTF) */
   /*{ VIA_RES_1024X576, CRTM1024x576, ARRAY_SIZE(CRTM1024x576)},*/

   /* Display : 1024x768 */
   { VIA_RES_1024X768, CRTM1024x768, ARRAY_SIZE(CRTM1024x768)},

   /* Display : 1152x864 */
   { VIA_RES_1152X864, CRTM1152x864, ARRAY_SIZE(CRTM1152x864)},

   /* Display : 1280x768 (GTF) */
   { VIA_RES_1280X768, CRTM1280x768, ARRAY_SIZE(CRTM1280x768)},

   /* Display : 1280x800 (GTF) */
   { VIA_RES_1280X800, CRTM1280x800, ARRAY_SIZE(CRTM1280x800)},

   /* Display : 1280x960 */
   { VIA_RES_1280X960, CRTM1280x960, ARRAY_SIZE(CRTM1280x960)},

   /* Display : 1280x1024 */
   { VIA_RES_1280X1024, CRTM1280x1024,ARRAY_SIZE(CRTM1280x1024)},

   /* Display : 1368x768 (GTF) */
   { VIA_RES_1368X768,CRTM1368x768,ARRAY_SIZE(CRTM1368x768)},

   /* Display : 1440x1050 (GTF) */
   { VIA_RES_1440X1050, CRTM1440x1050, ARRAY_SIZE(CRTM1440x1050)},

   /* Display : 1600x1200 */
   { VIA_RES_1600X1200, CRTM1600x1200, ARRAY_SIZE(CRTM1600x1200)},

   /* Display : 1920x1440 */
   { VIA_RES_1920X1440, CRTM1920x1440, ARRAY_SIZE(CRTM1920x1440)},

   /* Display : 1280x720 */
   { VIA_RES_1280X720, CRTM1280x720, ARRAY_SIZE(CRTM1280x720)},

   /* Display : 1920x1080 */
   { VIA_RES_1920X1080, CRTM1920x1080, ARRAY_SIZE(CRTM1920x1080)},

   /* Display : 1400x1050 */
   { VIA_RES_1400X1050, CRTM1400x1050, ARRAY_SIZE(CRTM1400x1050)}
};

#define NUM_TOTAL_MODETABLE ARRAY_SIZE(CLE266Modes)


#endif /* __VIAMODE_H__ */
