/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <linux/ioport.h>
#include <asm/io.h>

#include "tv.h"
#include "tbl1622.h"
#include "share.h"
#include "chip.h"
#include "debug.h"

extern inline void lock_crt(void);
extern inline void unlock_crt(void);
extern inline void write_reg(u8 index, u16 io_port, u8 data);
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern void load_crtc_timing(struct display_timing device_timing, int set_iga);
extern void write_tv_patch_table(u16 *tv_func_reg);

extern struct tv_setting_information    tv_setting_info;

int  vt1622_tv_encoder_identify(void)
{
    int ver;

    ver = tv_register_read(VT1622_VERSION_REG);
    if (ver == VT1622_VERSION)
    {
        DEBUG_MSG(KERN_INFO "\nFind VT1622 TV Encoder \n");
        return(OK);
    }
    else
        return(FAIL);
}

void vt1622_tv_disable(void)
{
    /* turn off TV's DAC */
    tv_register_write(0x0e, 0x0f);
}

void write_vt1622_patch_table(vt1622_func_table *func_table, int func_index)
{
    u16 *tv_func_reg = NULL;
    
    if (tv_setting_info.level == TV_SIZE_NORMAL_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_underscan;
    }
    else if (tv_setting_info.level == TV_SIZE_FIT_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_fitscan;
    }
    else if (tv_setting_info.level == TV_SIZE_OVER_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_overscan;
    }

    if (tv_func_reg != NULL)
    {
        write_tv_patch_table(tv_func_reg);
    }
}

void load_vt1622_regs(int TVModeIndex)
{
    int i, index;
    u8  tmp;
    vt1622_func_table            *vt1622_func;
    struct display_timing         tv_crtc_reg;
    u16                          *tv_reg=NULL;


    for(i=0;i<NUM_TOTAL_VT1622_TABLE;i++)
    {
       index = i;

       if (vt1622_tbl[i].tv_index == TVModeIndex)
       break;
    }

    switch (tv_setting_info.system)
    {
        case TVTYPE_NTSC:
        case TVTYPE_480P:
        {
            /* For 480P, we use NTSC's mode table, and then do some patches later. */
            vt1622_func = vt1622_tbl[index].tv_mode_ntsc;
            break;
        }
            
        case TVTYPE_PAL:
        case TVTYPE_576P:
        {
            /* For 576P, we use PAL's mode table, and then do some patches later. */
            vt1622_func = vt1622_tbl[index].tv_mode_pal;
            break;
        }

        default:
        {
            vt1622_func = vt1622_tbl[index].tv_mode_ntsc;
            break;
        }
    }

    if (tv_setting_info.level == TV_SIZE_NORMAL_SCAN)
        tv_reg = vt1622_func[0].tv_underscan;
    if (tv_setting_info.level == TV_SIZE_FIT_SCAN)
        tv_reg = vt1622_func[0].tv_fitscan;
    if (tv_setting_info.level == TV_SIZE_OVER_SCAN)
        tv_reg = vt1622_func[0].tv_overscan;



    /* unlock CRTC Register */
    unlock_crt();

    /* update starting address */
    write_reg(CR0C, VIACR, 0x00);
    write_reg(CR0D, VIACR, 0x00);
    /* write_reg(CR34, VIACR, 0x00); */
    write_reg_mask(CR48, VIACR, 0x00, BIT0+BIT1);
    /* Load TV Encoder Regsiters */
    for (i=0;i<NUM_TOTAL_VT1622_REG;i++)
    {
        tv_register_write(GET_LOW_BYTE(tv_reg[i]),GET_HIGH_BYTE(tv_reg[i]));
    }

    /* Load GFX CRTC timing when TV work on IGA1 */
    for (i=0; i< NUM_CRTC_TIMING; i++)
    {
        switch(i) {
            case H_TOTAL_INDEX:
                tv_crtc_reg.hor_total = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case H_ADDR_INDEX:
                tv_crtc_reg.hor_addr = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case H_BLANK_START_INDEX:
                tv_crtc_reg.hor_blank_start = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case H_BLANK_END_INDEX:
                tv_crtc_reg.hor_blank_end = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case H_SYNC_START_INDEX:
                tv_crtc_reg.hor_sync_start = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case H_SYNC_END_INDEX:
                tv_crtc_reg.hor_sync_end = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_TOTAL_INDEX:
                tv_crtc_reg.ver_total = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_ADDR_INDEX:
                tv_crtc_reg.ver_addr = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_BLANK_START_INDEX:
                tv_crtc_reg.ver_blank_start = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_BLANK_END_INDEX:
                tv_crtc_reg.ver_blank_end = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_SYNC_START_INDEX:
                tv_crtc_reg.ver_sync_start = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
            case V_SYNC_END_INDEX:
                tv_crtc_reg.ver_sync_end = tv_reg[i+NUM_TOTAL_VT1622_REG];
                break;
          }

    }

    if ( tv_setting_info.iga_path == IGA1)
        load_crtc_timing(tv_crtc_reg, IGA1);
    else
    {
        load_crtc_timing(tv_crtc_reg, IGA2);
        write_reg(CR65, VIACR, 0x80);
        write_reg(CR66, VIACR, 0x00);
        write_reg(CR67, VIACR, 0x41);
    }
    DEBUG_MSG(KERN_INFO "\n tv_out_signal = %2d \n", tv_setting_info.out_signal);

    /* Patch for RGB/YPbPr connector */
    if (tv_setting_info.out_signal & (TV_OUTPUT_RGB | TV_OUTPUT_RGB_COMPOSITE))
    {
        write_vt1622_patch_table(vt1622_func, TV_RGB_FUNC);
    }
    else if (tv_setting_info.out_signal & (TV_OUTPUT_YPBPR | TV_OUTPUT_YPBPR_COMPOSITE))
    {
        write_vt1622_patch_table(vt1622_func, TV_YPBPR_FUNC);
    }

    /* Patch for SDTV */
    if ((tv_setting_info.system == TVTYPE_480P) || (tv_setting_info.system == TVTYPE_576P))
    {
        if (tv_setting_info.out_signal == TV_OUTPUT_RGB)
        {
            write_vt1622_patch_table(vt1622_func, TV_SDTV_RGB_FUNC);
        }
        else if (tv_setting_info.out_signal == TV_OUTPUT_YPBPR)
        {
            write_vt1622_patch_table(vt1622_func, TV_SDTV_YPBPR_FUNC);
        }
    }
    
    if ((tv_setting_info.dedotcrawl == 1) && 
        (tv_setting_info.out_signal == TV_OUTPUT_COMPOSITE) && 
        (tv_setting_info.system == TVTYPE_NTSC))
    {
        tmp = (u8) tv_register_read(VT1622_DEDOTCRAWL_REG);
        tv_register_write(VT1622_DEDOTCRAWL_REG, tmp | BIT3);
        
        write_vt1622_patch_table(vt1622_func, TV_DeDotCrawl_FUNC);
    }

    lock_crt();
}

void vt1622_tv_enable(void)
{
    u8  data=0;

    data = (u8)tv_register_read(VT1622_DAC_REG);

    /* turn on TV's DAC */
    switch(tv_setting_info.out_signal)
    {
        case TV_OUTPUT_COMPOSITE:
            data = (data & 0xF0)|((~VT1622_DAC_A)&0x0F);
            break;
            
        case TV_OUTPUT_SVIDEO:
            data = (data & 0xF0)|((~(VT1622_DAC_B+VT1622_DAC_C))&0x0F);
            break;
            
        case TV_OUTPUT_COMPOSITE_SVIDEO:
            data = (data & 0xF0)|((~(VT1622_DAC_A+VT1622_DAC_B+VT1622_DAC_C))&0x0F);
            break;
            
        case TV_OUTPUT_RGB:
        case TV_OUTPUT_RGB_COMPOSITE:
        case TV_OUTPUT_YPBPR_COMPOSITE:
            data = (data & 0xF0)|((~(VT1622_DAC_A+VT1622_DAC_B+VT1622_DAC_C+VT1622_DAC_D))&0x0F);
            break;
            
        case TV_OUTPUT_YPBPR:
            data = (data & 0xF0)|((~(VT1622_DAC_B+VT1622_DAC_C+VT1622_DAC_D))&0x0F);
            break;

        default:
            data = (data & 0xF0)|((~(VT1622_DAC_B+VT1622_DAC_C+VT1622_DAC_D))&0x0F);

    }

    tv_register_write(VT1622_DAC_REG, data);
}

/* Sense VT1622 TV Encoder */
int vt1622_sense(void)
{
    u8  value, status = 0;
    int ret;

    DEBUG_MSG(KERN_INFO "vt1622_sense!!\n");

    value = (u8)tv_register_read(0x0E);
    tv_register_write(0x0E, 0x0);
    tv_register_write(0x0E, 0x80);
    tv_register_write(0x0E, 0x0);
    status = (u8)tv_register_read(0x0F);
    tv_register_write(0x0E, value);

    switch (status&0x0F)
    {
        case 0x0:   /* DAC A, B, C, D */
            ret = TV_OUTPUT_RGB;
            break;
        case 0x01:  /* DAC A, B, C */
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
        case 0x07:  /* DAC A */
            ret = TV_OUTPUT_COMPOSITE;
            break;
        case 0x08:  /* DAC B, C, D */
            ret = TV_OUTPUT_YPBPR;
            break;
        case 0x09:  /* DAC B, C */
            ret = TV_OUTPUT_SVIDEO;
            break;
        case 0x0F:
            ret = TV_OUTPUT_NONE;
            break;
        default:
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
    }
    
    return ret;
}

/* Auto-sense VT1622 TV Encoder */
int vt1622_auto_sense(void)
{
    u8  status = 0x0;
    int ret;

    DEBUG_MSG(KERN_INFO "vt1622_auto_sense!!\n");

    status = (u8)tv_register_read(0x0F);

    switch (status&0x0F)
    {
        case 0x0:   /* DAC A, B, C, D */
            ret = TV_OUTPUT_RGB;
            break;
        case 0x01:  /* DAC A, B, C */
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
        case 0x07:  /* DAC A */
            ret = TV_OUTPUT_COMPOSITE;
            break;
        case 0x08:  /* DAC B, C, D */
            ret = TV_OUTPUT_YPBPR;
            break;
        case 0x09:  /* DAC B, C */
            ret = TV_OUTPUT_SVIDEO;
            break;
        case 0x0F:
            ret = TV_OUTPUT_NONE;
            break;
        default:
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
    }

    return ret;
}

