/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <linux/ioport.h>
#include <asm/io.h>

#include "tv.h"
#include "tbl1625.h"
#include "share.h"
#include "chip.h"
#include "debug.h"

extern inline void lock_crt(void);
extern inline void unlock_crt(void);
extern inline void write_reg(u8 index, u16 io_port, u8 data);
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern void load_crtc_timing(struct display_timing device_timing, int set_iga);
extern void write_tv_patch_table(u16 *tv_func_reg);
extern void plus_vck_to_iga1_timing(unsigned long h_total);

extern struct tv_setting_information    tv_setting_info;
extern struct chip_information          chip_info;

int  vt1625_tv_encoder_identify(void)
{
    int ver;

    ver = tv_register_read(VT1625_VERSION_REG);
    if (ver == VT1625_VERSION)
    {
        DEBUG_MSG(KERN_INFO "\nFind VT1625 TV Encoder \n");
        return(OK);
    }
    else
        return(FAIL);
}

void vt1625_tv_disable(void)
{
    /* turn off TV's DAC */
    tv_register_write(0x0E, 0x3F);
    /* reset TV */
    tv_register_write(0x1D, 0x00);
    /* disable reset TV */
    tv_register_write(0x1D, 0x80);
}

void write_vt1625_patch_table(vt1625_func_table *func_table, int func_index)
{
    u16 *tv_func_reg = NULL;
    
    if (tv_setting_info.level == TV_SIZE_NORMAL_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_underscan;
    }
    else if (tv_setting_info.level == TV_SIZE_FIT_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_fitscan;
    }
    else if (tv_setting_info.level == TV_SIZE_OVER_SCAN)
    {
        tv_func_reg = func_table[func_index].tv_overscan;
    }

    if (tv_func_reg != NULL)
    {
        write_tv_patch_table(tv_func_reg);
    }
}

void load_vt1625_regs(int TVModeIndex)
{
    int i, index;
    u8  tmp;
    vt1625_func_table            *vt1625_func;
    struct display_timing         tv_crtc_reg;
    u16                          *tv_reg = NULL;

    for(i=0;i<NUM_TOTAL_VT1625_TABLE;i++)
    {
        index = i;

        if(vt1625_tbl[i].tv_index == TVModeIndex)
            break;
    }

    switch (tv_setting_info.system)
    {
        case TVTYPE_NTSC:
        case TVTYPE_480P:
        {
            /* For 480P, we use NTSC's mode table, and then do some patches later. */
            vt1625_func = vt1625_tbl[index].tv_mode_ntsc;
            break;
        }
            
        case TVTYPE_PAL:
        case TVTYPE_576P:
        {
            /* For 576P, we use PAL's mode table, and then do some patches later. */
            vt1625_func = vt1625_tbl[index].tv_mode_pal;
            break;
        }

        case TVTYPE_720P:
        {
            vt1625_func = vt1625_tbl[index].tv_mode_720p;
            break;
        }

        case TVTYPE_1080I:
        {
            vt1625_func = vt1625_tbl[index].tv_mode_1080i;
            break;
        }

        default:
        {
            vt1625_func = vt1625_tbl[index].tv_mode_ntsc;
            break;
        }
    }


    if (tv_setting_info.level == TV_SIZE_NORMAL_SCAN)
    {
        tv_reg = vt1625_func[0].tv_underscan;
    }
    
    if (tv_setting_info.level == TV_SIZE_FIT_SCAN)
    {
        tv_reg = vt1625_func[0].tv_fitscan;
    }
    
    if (tv_setting_info.level == TV_SIZE_OVER_SCAN)
    {
        tv_reg = vt1625_func[0].tv_overscan;
    }

    /* unlock CRTC Register */
    unlock_crt();

    /* update starting address */
    write_reg(CR0C, VIACR, 0x00);
    write_reg(CR0D, VIACR, 0x00);

    write_reg_mask(CR48, VIACR, 0x00, BIT0+BIT1);

    /* Load TV Encoder Regsiters */
    for (i=0;i<NUM_TOTAL_VT1625_REG;i++)
    {
        tv_register_write(GET_LOW_BYTE(tv_reg[i]),GET_HIGH_BYTE(tv_reg[i]));
    }

    /* Load GFX CRTC timing when TV work on IGA1 */
    for (i=0; i< NUM_CRTC_TIMING; i++)
    {
        switch(i)
        {
            case H_TOTAL_INDEX:
                tv_crtc_reg.hor_total = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case H_ADDR_INDEX:
                tv_crtc_reg.hor_addr = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case H_BLANK_START_INDEX:
                tv_crtc_reg.hor_blank_start = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case H_BLANK_END_INDEX:
                tv_crtc_reg.hor_blank_end = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case H_SYNC_START_INDEX:
                tv_crtc_reg.hor_sync_start = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case H_SYNC_END_INDEX:
                tv_crtc_reg.hor_sync_end = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_TOTAL_INDEX:
                tv_crtc_reg.ver_total = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_ADDR_INDEX:
                tv_crtc_reg.ver_addr = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_BLANK_START_INDEX:
                tv_crtc_reg.ver_blank_start = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_BLANK_END_INDEX:
                tv_crtc_reg.ver_blank_end = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_SYNC_START_INDEX:
                tv_crtc_reg.ver_sync_start = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
            case V_SYNC_END_INDEX:
                tv_crtc_reg.ver_sync_end = tv_reg[i+NUM_TOTAL_VT1625_REG];
                break;
        }
    }

    if (tv_setting_info.iga_path == IGA1)
    {
        load_crtc_timing(tv_crtc_reg, IGA1);

        /* patch TV IGA1 horizontal total can't be divided by 8 */        
        if ((chip_info.tv_chip_info.tv_chip_name == VT1625) && 
            ((chip_info.gfx_chip_name == UNICHROME_P880) || (chip_info.gfx_chip_name == UNICHROME_CN900)))
        {
            plus_vck_to_iga1_timing(tv_crtc_reg.hor_total);
        }
    }
    else
    {
        load_crtc_timing(tv_crtc_reg, IGA2);
        write_reg(CR65, VIACR, 0x80);
        write_reg(CR66, VIACR, 0x00);
        write_reg(CR67, VIACR, 0x41);
    }

    /* Load CGMS registers */
    for (i=0;i<NUM_TOTAL_VT1625_CGMS_TBL;i++)
    {
        tv_register_write(GET_LOW_BYTE(VT1625_CGMS_TBL[i]),GET_HIGH_BYTE(VT1625_CGMS_TBL[i]));
    }

    DEBUG_MSG(KERN_INFO "\n tv_out_signal = %2d \n", tv_setting_info.out_signal);

    /* Patch for RGB/YPbPr connector */
    if (tv_setting_info.out_signal & (TV_OUTPUT_RGB | TV_OUTPUT_RGB_COMPOSITE | TV_OUTPUT_RGB_COMPOSITE_SVIDEO))
    {
        write_vt1625_patch_table(vt1625_func, TV_RGB_FUNC);
    }
    else if (tv_setting_info.out_signal & (TV_OUTPUT_YPBPR | TV_OUTPUT_YPBPR_COMPOSITE | TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO))
    {
        write_vt1625_patch_table(vt1625_func, TV_YPBPR_FUNC);
    }

    /* Patch for SDTV */
    if ((tv_setting_info.system == TVTYPE_480P) || (tv_setting_info.system == TVTYPE_576P))
    {
        if (tv_setting_info.out_signal == TV_OUTPUT_RGB)
        {
            write_vt1625_patch_table(vt1625_func, TV_SDTV_RGB_FUNC);
        }
        else if (tv_setting_info.out_signal == TV_OUTPUT_YPBPR)
        {
            write_vt1625_patch_table(vt1625_func, TV_SDTV_YPBPR_FUNC);
        }
    }

    /* Patch for DeDotCrawl function */
    if ((tv_setting_info.dedotcrawl == 1) && 
        (tv_setting_info.out_signal== TV_OUTPUT_COMPOSITE) && 
        (tv_setting_info.system == TVTYPE_NTSC))
    {
        tmp = (u8) tv_register_read(VT1625_DEDOTCRAWL_REG);
        tv_register_write(VT1625_DEDOTCRAWL_REG, tmp | BIT3);
        
        write_vt1625_patch_table(vt1625_func, TV_DeDotCrawl_FUNC);
    }
     
    lock_crt();
}

void vt1625_tv_enable(void)
{
    u8  data;

    data = (u8)tv_register_read(VT1625_DAC_REG);

    /*  turn on TV's DAC */
    switch(tv_setting_info.out_signal)
    {
        case TV_OUTPUT_COMPOSITE:
            data = (data & 0xC0)|((~VT1625_DAC_C)&0x3F);
            break;
        case TV_OUTPUT_SVIDEO:
            data = (data & 0xC0)|((~(VT1625_DAC_A+VT1625_DAC_B))&0x3F);
            break;
        case TV_OUTPUT_COMPOSITE_SVIDEO:
            data = (data & 0xC0)|((~(VT1625_DAC_A+VT1625_DAC_B+VT1625_DAC_C))&0x3F);
            break;
        case TV_OUTPUT_RGB_COMPOSITE:
        case TV_OUTPUT_YPBPR_COMPOSITE:
            data = (data & 0xC0)|((~(VT1625_DAC_C+VT1625_DAC_D+VT1625_DAC_E+VT1625_DAC_F))&0x3F);
            break;
        case TV_OUTPUT_RGB:
        case TV_OUTPUT_YPBPR:
            data = (data & 0xC0)|((~(VT1625_DAC_D+VT1625_DAC_E+VT1625_DAC_F))&0x3F);
            break;
        case TV_OUTPUT_RGB_COMPOSITE_SVIDEO:
        case TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO:
            data = (data & 0xC0)|((~(VT1625_DAC_A+VT1625_DAC_B+VT1625_DAC_C+VT1625_DAC_D+VT1625_DAC_E+VT1625_DAC_F))&0x3F);
            break;
        default:
            data = (data & 0xC0)|((~(VT1625_DAC_A+VT1625_DAC_B+VT1625_DAC_C))&0x3F);
            break;
    }

    tv_register_write(VT1625_DAC_REG, data);
}


/* Sense VT1625 TV Encoder */
int vt1625_sense(void)
{
    u8  Reg_TV0E, Reg_TV1C,  Reg_TV0F = 0;
    int ret;

    DEBUG_MSG(KERN_INFO "vt1625_sense!!\n");

    /* Fill sense data */
    tv_register_write(0x69, 0x90);
    tv_register_write(0x6A, 0x02);

    Reg_TV0E = (u8)tv_register_read(0x0E);
    tv_register_write(0x0E, 0x0);
    Reg_TV1C = (u8)tv_register_read(0x1C);
    tv_register_write(0x1C, 0x80);
    tv_register_write(0x1C, Reg_TV1C);
    Reg_TV0F = (u8)tv_register_read(0x0F);
    tv_register_write(0x0E, Reg_TV0E);

    switch (Reg_TV0F&0x3F)
    {
        case 0x03:  /* DAC A, B, C, D (?)*/
            ret = TV_OUTPUT_RGB;
            break;
        case 0x07:  /* DAC A, B, C */
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
        case 0x0F:  /* DAC A, B */
            ret = TV_OUTPUT_SVIDEO;
            break;
        case 0x38:  /* DAC D, E, F */
            ret = TV_OUTPUT_YPBPR;
            break;
        case 0x37:  /* DAC C */
            ret = TV_OUTPUT_COMPOSITE;
            break;
        case 0x0:  /* DAC A, B, C, D, E, F */
        case 0x10:  /* S-Video C may not be detected. */
            ret = TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO;
            break;
        case 0x3F:
            ret = TV_OUTPUT_NONE;
            break;
        default:
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
    }
    
    return ret;
}

/* Auto-sense VT1625 TV Encoder */
int vt1625_auto_sense(void)
{
    u8  reg_tv1c, status = 0x0;
    int ret;

    DEBUG_MSG(KERN_INFO "vt1625_auto_sense!!\n");

    /* Enable TV auto sense. */
    reg_tv1c = (u8)tv_register_read(0x1C) | 0x40;
    tv_register_write(0x1C, reg_tv1c);

    /* Fill sense data */
    tv_register_write(0x69, 0x90);
    tv_register_write(0x6A, 0x02);

    status = (u8)tv_register_read(0x0F);

    /* Disable TV auto sense. */
    reg_tv1c = (u8)tv_register_read(0x1C) & 0xBF;
    tv_register_write(0x1C, reg_tv1c);

    switch (status&0x3F)
    {
        case 0x03:  /* DAC A, B, C, D (?)*/
            ret = TV_OUTPUT_RGB;
            break;
        case 0x07:  /* DAC A, B, C */
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
        case 0x0F:  /* DAC A, B */
            ret = TV_OUTPUT_SVIDEO;
            break;
        case 0x38:  /* DAC D, E, F */
            ret = TV_OUTPUT_YPBPR;
            break;
        case 0x37:  /* DAC C */
            ret = TV_OUTPUT_COMPOSITE;
            break;
        case 0x0:  /* DAC A, B, C, D, E, F */
        case 0x10:  /* S-Video C may not be detected. */
            ret = TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO;
            break;
        case 0x3F:
            ret = TV_OUTPUT_NONE;
            break;
        default:
            ret = TV_OUTPUT_COMPOSITE_SVIDEO;
            break;
    }

    return ret;
}

void init_tv_vt1625(void)
{
    /* TV 5 init value */
    tv_register_write(0x4A, 0xC5);   /* DAC A/B/C/D select */
    tv_register_write(0x4B, 0x0F);   /* DAC E/F select */
    tv_register_write(0x69, 0x90);   /* init sense data */
    tv_register_write(0x6A, 0x02);   /* init sense data */
    tv_register_write(0x1C, 0);
}
bool is_vt1625_mode(struct tv_chip_information* ptv_chip_info, struct tv_setting_information* ptv_setting_info,
                             int video_index)
{
    vt1625_func_table    *vt1625_func;
    u16  *tv_reg = NULL;
    int i, index;
    if (VIA_RES_640X480 == video_index ||VIA_RES_800X600 == video_index ||
        VIA_RES_1024X768 == video_index || VIA_RES_720X480 == video_index ||
        VIA_RES_720X576 == video_index || VIA_RES_1280X720 == video_index ||
        VIA_RES_1920X1080 == video_index)
        {
            for(i=0;i<NUM_TOTAL_VT1625_TABLE;i++)
            {
                index = i;

                if(vt1625_tbl[i].tv_index == video_index)
                    break;
            }

            switch (ptv_setting_info->system)
            {
                case TVTYPE_NTSC:
                case TVTYPE_480P:
                {
                    /* For 480P, we use NTSC's mode table, and then do some patches later. */
                    vt1625_func = vt1625_tbl[index].tv_mode_ntsc;
                    break;
                }
                    
                case TVTYPE_PAL:
                case TVTYPE_576P:
                {
                    /* For 576P, we use PAL's mode table, and then do some patches later. */
                    vt1625_func = vt1625_tbl[index].tv_mode_pal;
                    break;
                }

                case TVTYPE_720P:
                {
                    vt1625_func = vt1625_tbl[index].tv_mode_720p;
                    break;
                }

                case TVTYPE_1080I:
                {
                    vt1625_func = vt1625_tbl[index].tv_mode_1080i;
                    break;
                }

                default:
                {
                    vt1625_func = vt1625_tbl[index].tv_mode_ntsc;
                    break;
                }
            }
            if(IS_VT1625_NULL_TBL(vt1625_func))
            {
                WARN_MSG(KERN_WARNING "The specified tv mode is not valid under specified tv system\n");
                return FALSE;
            }
            if (ptv_setting_info->level == TV_SIZE_NORMAL_SCAN)
            {
                tv_reg = vt1625_func[0].tv_underscan;
            }
            
            if (ptv_setting_info->level == TV_SIZE_FIT_SCAN)
            {
                tv_reg = vt1625_func[0].tv_fitscan;
            }
            
            if (ptv_setting_info->level == TV_SIZE_OVER_SCAN)
            {
                tv_reg = vt1625_func[0].tv_overscan;
            }
            if(IS_VT1625_NULL_FUNC(tv_reg))
            {
                WARN_MSG(KERN_WARNING "The specified tv level is not valid under spcified combination of tv mode and system\n");
                return FALSE;
            }
            return TRUE;
        }
        WARN_MSG(KERN_WARNING "The specified tv mode is not valid for currently used  tv encoder\n");
        return FALSE;
}
