/*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include "hw.h"
#include "tbl1636.h"
#include "vt1636.h"
#include "tblDPASetting.h"
#include "chip.h"
#include "via_i2c.h"
#include "debug.h"

/* Extern Functions Prototypes Declaration */
extern int i2cWriteByte(u8 slave_addr, u8 index, u8 data);
extern int i2cReadByte(u8 slave_addr, u8 index, u8 *pData);
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern void set_dpa_gfx(int output_interface, GFXDPASETTINGPTR p_gfx_dpa_setting);
extern struct chip_information      chip_info;


u8 gpio_i2c_read_lvds(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info, u8 index)
{
    u8 data;
    int tmp;
    tmp = chip_info.tv_chip_info.tv_chip_on_slot;
    switch(plvds_chip_info->i2c_port)
    {
        case TVI2CPORTINDEX:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AMR;
            break;
        case TVGPIOPORTINDEX:
        default:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AGP;
            break;
    }
    i2cReadByte(plvds_chip_info->lvds_chip_slave_addr, index, &data);
    chip_info.tv_chip_info.tv_chip_on_slot = tmp;
    return data;
}

void gpio_i2c_write_mask_lvds(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info, IODATA io_data)
{
    int index, data, tmp;
    tmp = chip_info.tv_chip_info.tv_chip_on_slot;
    switch(plvds_chip_info->i2c_port)
    {
        case TVI2CPORTINDEX:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AMR;
            break;
        case TVGPIOPORTINDEX:
        default:
            chip_info.tv_chip_info.tv_chip_on_slot = TV_ON_AGP;
            break;
    }

    index = io_data.Index;
    data = gpio_i2c_read_lvds(plvds_setting_info, plvds_chip_info, index);
    data = (data & (~io_data.Mask)) | io_data.Data;

    i2cWriteByte(plvds_chip_info->lvds_chip_slave_addr, index, data);
    chip_info.tv_chip_info.tv_chip_on_slot = tmp;
}

void init_lvds_vt1636(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    int reg_num, i;

    /* Common settings: */
    reg_num = ARRAY_SIZE(COMMON_INIT_TBL_VT1636);

    for (i = 0; i < reg_num; i++)
    {
        gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, COMMON_INIT_TBL_VT1636[i]);
    }

    /* Input Data Mode Select */
    if (plvds_setting_info->LCDDualEdge)
    {
        gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, DUAL_CHANNEL_ENABLE_TBL_VT1636[0]);
    }
    else
    {
        gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, SINGLE_CHANNEL_ENABLE_TBL_VT1636[0]);
    }

    if (plvds_setting_info->LCDDithering)
    {
        gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, DITHERING_ENABLE_TBL_VT1636[0]);
    }
    else
    {
        gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, DITHERING_DISABLE_TBL_VT1636[0]);
    }
}


void enable_lvds_vt1636(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    
    gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, VDD_ON_TBL_VT1636[0]);

    /* Pad on: */
    switch (plvds_chip_info->output_interface)
    {
        case INTERFACE_DVP0:
        {
            write_reg_mask(SR1E, VIASR, 0xC0, 0xC0);
            break;
        }
        
        case INTERFACE_DVP1:
        {
            write_reg_mask(SR1E, VIASR, 0x30, 0x30);
            break;
        }

        case INTERFACE_DFP_LOW:
        {
            write_reg_mask(SR2A, VIASR, 0x03, 0x03);
            break;
        }

        case INTERFACE_DFP_HIGH:
        {
            write_reg_mask(SR2A, VIASR, 0x03, 0x0C);
            break;
        }
       
        
    }
}


void disable_lvds_vt1636(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    
    gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, VDD_OFF_TBL_VT1636[0]);

    /* Pad off: */
    switch (plvds_chip_info->output_interface)
    {
        case INTERFACE_DVP0:
        {
            write_reg_mask(SR1E, VIASR, 0x00, 0xC0);
            break;
        }

        case INTERFACE_DVP1:
        {
            write_reg_mask(SR1E, VIASR, 0x00, 0x30);
            break;
        }

        case INTERFACE_DFP_LOW:
        {
            write_reg_mask(SR2A, VIASR, 0x00, 0x03);
            break;
        }

        case INTERFACE_DFP_HIGH:
        {
            write_reg_mask(SR2A, VIASR, 0x00, 0x0C);
            break;
        }
        
    }
}


bool lvds_identify_vt1636(void)
{
    u8 Buffer[2];

    DEBUG_MSG(KERN_INFO "lvds_identify_vt1636.\n");
    
    /* Sense VT1636 LVDS Transmiter */
    chip_info.lvds_chip_info.lvds_chip_slave_addr = VT1636_LVDS_I2C_ADDR ;
    
    /* Check vendor ID first: */
    i2cReadByte((u8)chip_info.lvds_chip_info.lvds_chip_slave_addr, 0x00, &Buffer[0]);
    i2cReadByte((u8)chip_info.lvds_chip_info.lvds_chip_slave_addr, 0x01, &Buffer[1]);

    if (!((Buffer[0] == 0x06) && (Buffer[1] == 0x11)))
    {      
        return FALSE;
    }

    /* Check Chip ID: */
    i2cReadByte((u8)chip_info.lvds_chip_info.lvds_chip_slave_addr, 0x02, &Buffer[0]);
    i2cReadByte((u8)chip_info.lvds_chip_info.lvds_chip_slave_addr, 0x03, &Buffer[1]);
    if ((Buffer[0] == 0x45) && (Buffer[1] == 0x33))
    {
        chip_info.lvds_chip_info.lvds_chip_name = VT1636_LVDS;
        return TRUE;
    }
    
    return FALSE;
}



int get_clk_range_index(u32 Clk)
{
    if (Clk < DPA_CLK_30M)
    {
        return DPA_CLK_RANGE_30M;
    }
    else if ((DPA_CLK_30M < Clk) && (Clk < DPA_CLK_50M))
    {
        return DPA_CLK_RANGE_30_50M;
    }
    else if ((DPA_CLK_50M < Clk) && (Clk < DPA_CLK_70M))
    {
        return DPA_CLK_RANGE_50_70M;
    }
    else if ((DPA_CLK_70M < Clk) && (Clk < DPA_CLK_100M))
    {
        return DPA_CLK_RANGE_70_100M;
    }
    else if ((DPA_CLK_100M < Clk) && (Clk < DPA_CLK_150M))
    {
        return DPA_CLK_RANGE_100_150M;
    }
    else
    {
        return DPA_CLK_RANGE_150M;
    }
}


int get_lvds_dpa_setting_index(int panel_size_id, VT1636DPASETTINGPTR p_vt1636_dpasetting_tbl, int tbl_size)
{
    int i;

    for (i = 0; i < tbl_size; i++)
    {
        if (panel_size_id == p_vt1636_dpasetting_tbl->PanelSizeID)
        {
            return i;
        }

        p_vt1636_dpasetting_tbl++;
    }

    return 0;
}


void set_dpa_vt1636(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info, VT1636DPASETTINGPTR p_vt1636_dpa_setting)
{
    IODATA io_data;

    io_data.Index = 0x09;
    io_data.Mask = 0x1F;
    io_data.Data = p_vt1636_dpa_setting->CLK_SEL_ST1;
    gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, io_data);

    io_data.Index = 0x08;
    io_data.Mask = 0x0F;
    io_data.Data = p_vt1636_dpa_setting->CLK_SEL_ST2;
    gpio_i2c_write_mask_lvds(plvds_setting_info, plvds_chip_info, io_data);
}


void vt1636_patch_skew_on_vt3324(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    int index, size;

    DEBUG_MSG(KERN_INFO "vt1636_patch_skew_on_vt3324.\n");
    
    /* Graphics DPA settings: */
    index = get_clk_range_index(plvds_setting_info->vclk);
    set_dpa_gfx(plvds_chip_info->output_interface, &GFX_DPA_SETTING_TBL_VT3324[index]);

    /* LVDS Transmitter DPA settings: */
    size = ARRAY_SIZE(VT1636_DPA_SETTING_TBL_VT3324);
    index = get_lvds_dpa_setting_index(plvds_setting_info->lcd_panel_id, VT1636_DPA_SETTING_TBL_VT3324, size);
    set_dpa_vt1636(plvds_setting_info, plvds_chip_info, &VT1636_DPA_SETTING_TBL_VT3324[index]);
}


void vt1636_patch_skew_on_vt3327(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    int index, size;

    DEBUG_MSG(KERN_INFO "vt1636_patch_skew_on_vt3327.\n");
    
    /* Graphics DPA settings: */
    index = get_clk_range_index(plvds_setting_info->vclk);
    set_dpa_gfx(plvds_chip_info->output_interface, &GFX_DPA_SETTING_TBL_VT3327[index]);

    /* LVDS Transmitter DPA settings: */
    size = ARRAY_SIZE(VT1636_DPA_SETTING_TBL_VT3327);
    index = get_lvds_dpa_setting_index(plvds_setting_info->lcd_panel_id, VT1636_DPA_SETTING_TBL_VT3327, size);
    set_dpa_vt1636(plvds_setting_info, plvds_chip_info, &VT1636_DPA_SETTING_TBL_VT3327[index]);
}

void vt1636_patch_skew_on_vt3364(struct lvds_setting_information *plvds_setting_info, struct lvds_chip_information *plvds_chip_info)
{
    int index;

    DEBUG_MSG(KERN_INFO "vt1636_patch_skew_on_vt3364.\n");
    
    /* Graphics DPA settings: */
    index = get_clk_range_index(plvds_setting_info->vclk);
    set_dpa_gfx(plvds_chip_info->output_interface, &GFX_DPA_SETTING_TBL_VT3364[index]);
}

